import pathlib
from queue import Empty
import multiprocessing as mp
    

from piidigger import classes
from piidigger import globalfuncs
from piidigger import queuefuncs
from piidigger.getmime import getMime
from piidigger.globalvars import SENTINEL
from piidigger.logmanager import LogManager

try:
    from win32api import GetFileAttributes
    win32apiLoaded=True
except ModuleNotFoundError:
    win32apiLoaded=False


def findDirsWorker(config: classes.Config, 
                   queues: dict, 
                   totals: dict,
                   stopEvent: mp.Event,
                   logManager: LogManager,) -> list:
    '''
    Expects a config object to start from (to get the startDirs paths)

    Populates a queue of all directories that aren't on the excludedDirs list
    '''
    
    
    try:
        ctrlc=False
        logger = logManager.getLogger('findDirsWorker')        
        
        logger.info('Starting findDirsWorker')

        localQ=list()
        i=0

        for d in config.getStartDirs():
            localQ.append(pathlib.Path(d))
            queues['dirsQ'].put(pathlib.Path(d))
            with totals['dirsFound'].get_lock():
                totals['dirsFound'].value+=1

        while not stopEvent.is_set() or not ctrlc:
            try:
                p=localQ[i]
            except IndexError:
                break
            else:
                i+=1

            
            try:
                for subD in p.iterdir():
                    if stopEvent.is_set():
                        break
                    excludeDir=False
                    try:
                        if subD.is_dir() and not subD.is_symlink():
                            for pattern in config.getExcludeDirs():
                                if str(subD).lower().startswith(pattern.lower()):
                                    logger.debug('Excluding directory %s matched pattern %s', str(subD), pattern)
                                    excludeDir=True
                                    break
                            if not excludeDir:
                                logger.debug('Including directory %s', str(subD))
                                localQ.append(subD)
                                queues['dirsQ'].put(subD)
                                with totals['dirsFound'].get_lock():
                                    totals['dirsFound'].value+=1
                    except FileNotFoundError:
                        pass
                    except OSError as e:
                        logger.debug('OSError: %s', str(e))
            except PermissionError as e:
                logger.debug('PermissionError: %s', str(e))
            except FileNotFoundError as e:
                logger.debug('FileNotFoundError: %s', str(e))
        
        # All directories have been scanned.  Put the sentinel on the queue for the fileHandlers
        queues['dirsQ'].put(SENTINEL)
        queuefuncs.waitOnQ(queues['dirsQ'])
    except KeyboardInterrupt:
        logger.info('KeyboardInterrupt received in findDirsWorker')
        queuefuncs.clearQ(queues['dirsQ'])
    finally:
        # All directories have been scanned.  Send the sentinel message to shutdown the consumer threads
        logger.info('Found %d folders', totals['dirsFound'].value)
        logger.info('Stopping %s (PID=%d)', mp.current_process().name, mp.current_process().pid)

        
def findFilesWorker(config: classes.Config, 
                    queues: dict, 
                    totals: dict,
                    stopEvent: mp.Event,
                    logManager: LogManager,) -> list:
    '''
    Inputs: Config(config), dirsQ as consumer, filesQ as producer

    Places identified files on the filesQ for one of the file handler workers to pick it up
    '''
    
    
    try:
        logger = logManager.getLogger(name=mp.current_process().name,)
        
        logger.info('Starting %s', mp.current_process().name)

        while True:
            if stopEvent.is_set():
                break
            item=queuefuncs.getItem(queues['dirsQ'])
            if item == SENTINEL:
                break
            if item == None:
                continue

            # Path-ify the directory name
            d=pathlib.Path(item)
            with totals['dirsScanned'].get_lock():
                totals['dirsScanned'].value+=1
            
            try: 
                logger.info('Scanning directory: %s', str(d))
                for f in d.iterdir():
                    screenItem=fileChecks(f, config)           
                    if all(screenItem):
                        mimeType = getMime(f)
                        match=fileMatches(f, config.getFileExts(), config.getMimeTypes())
                        if match:
                            fObj=classes.File(f, mimeType)
                            logger.debug('Initialized File object for %s, mimeType=%s, times=%s, handler=%s', 
                                         fObj.getFullPath(), 
                                         mimeType, 
                                         str(fObj.getTimeStamps()), 
                                         fObj.getFileHandlerName())
                            queues['filesQ'].put(fObj)
                            with totals['filesFound'].get_lock():
                                totals['filesFound'].value+=1
                            with totals['bytesFound'].get_lock():
                                totals['bytesFound'].value+=fObj.getFileSize()
                        else:
                            logger.debug('%s: Item not added (suffix: %s | mime: %s)', f, f.suffix, mimeType)
                    else:
                        logger.debug('%s: Item failed file checks (isFile=%s, isNotZero=%s, isLocalFile=%s)', f, screenItem[0], screenItem[1], screenItem[2])
            except PermissionError as e:
                logger.debug('PermissionError: %s', str(e))
            except FileNotFoundError:
                pass
            except OSError as e:
                logger.debug('OSError: %s', str(e))

        # All files have been identified.  Put the sentinel on the queue for the fileHandlers 
        queues['filesQ'].put(SENTINEL)
        queuefuncs.waitOnQ(queues['filesQ'])
    except KeyboardInterrupt:
        logger.info('KeyboardInterrupt received in findFilesWorker')
        queuefuncs.clearQ(queues['dirsQ'])
        queuefuncs.clearQ(queues['filesQ'])
    finally:
        logger.info('Found %d files', totals['filesFound'].value)
        logger.info('Stopping %s (PID=%d)', mp.current_process().name, mp.current_process().pid)


def fileChecks(f: pathlib.Path, 
               config: classes.Config) -> tuple:
    '''
    Performs initial checks to see if the item is a file, has content, and is local (e.g. OneDrive file already local file system)
    Returns a tuple of (isFile, isNotZero, isLocalFile)
    '''
    recallMask=0x400000
    offlineMask=0x1000

    isFile = f.is_file() 
    isNotZero = f.stat().st_size > 0

    # Checks if the file is local.  Non-local OneDrive and Dropbox files will be skipped.
    # https://superuser.com/questions/1718444/determining-if-a-onedrive-file-is-synced-locally-via-a-terminal
    # For OneDrive, the 0x400000 (Recall) bit will be set (TRUE) if the file isn't on the disk.
    # For Dropbox, the 0x1000 (Offline) bit will be set (TRUE) if the file isn't on the disk.
    # According to the linked article, this might not be derived from MS documentation, so it could break in the future, but from my own testing, it seems to indicate correctly
    # And the Dropbox behavior is from direct observation
    if win32apiLoaded and config.getLocalFilesOnly():
        try:
            attr=GetFileAttributes(str(f))
        except Exception:
            # The most common exception is that the file is in use, which means its local. But, when in doubt, assume it's a local file.
            isLocalFile=True
        else:
            oneDriveRemote=bool(attr & recallMask)
            dropboxRemote=bool(attr & offlineMask)

            # A file is only remote when either the RECALL or the Offline bit is set.  We need the inverse of that.
            isLocalFile=not(oneDriveRemote or dropboxRemote)
    else:
        isLocalFile=True

    return (isFile, isNotZero, isLocalFile)

def fileMatches(f: pathlib.Path, fileExts: list, mimeTypes: list) -> bool:
    extFound = f.suffix in fileExts
    mimeFound = getMime(f) in mimeTypes

    return extFound or mimeFound






import re

dhName='Primary Account Number'

def findMatch(line: str) -> dict:
    '''
    Matches a line of text against known credit card number formats.  Should receive the text from "filehandler" as raw text.

    Returns a dictionary of:
        'brand': set(matches)
    '''

    # Regexes provided by https://github.com/citypay/citypay-pan-search/tree/master/src/test/resources
    
    # If you know of more, or can shed light on any corrections, please submit an issue at https://github.com/kirkpatrickprice/PIIDigger/issues or submit a PR on the repo
    # Added the |[^-] to exclude strings anchored on a hyphen.  Hopefully reduce UUID false positives in log files without missing legit PAN.
    regexes={
        'visa': re.compile(r'(?:^|[^\d.-])4[0-9]{3}[ -]?[0-9]{4}[ -]?[0-9]{4}[ -]?[0-9]{4}(?:$|[^\d.-])'),
        'mc': re.compile(r'(?:^|[^\d.-])5[1-5][0-9]{2}[ -]?[0-9]{4}[ -]?[0-9]{4}[ -]?[0-9]{4}(?:$|[^\d.-])'),
        'discover': re.compile(r'(?:^|[^\d.-])6011[ -]?[0-9]{4}[ -]?[0-9]{4}[ -]?[0-9]{4}(?:$|[^\d.-])'),
        'jcb': re.compile(r'(?:^|[^\d.-])(?:2131|1800|35[0-9]{3})[0-9]{11}(?:$|[^\d.-])'),
        'amex': re.compile(r'(?:^|[^\d.-])3[47][0-9]{2}[ -]?[0-9]{6}[ -]?[0-9]{5}(?:$|[^\d.-])'),
        }
    
    results=dict()
    for brand in regexes.keys():
        matches=regexes[brand].findall(line)
        if matches:
            for match in matches:
                match=match.strip()
                if _isValid(match):
                    results[brand]=set()
                    try:
                        results[brand].add(_redact(match))
                    except KeyError:
                        results[brand]=[]
                        results[brand]+=[_redact(match)]
    return results


def _isValid(text: str) -> bool:

    def luhn(n) -> bool:
        # Luhn check taken from https://rosettacode.org/wiki/Luhn_test_of_credit_card_numbers#Python

        # Invert the number string for the Luhn formula
        r = [int(ch) for ch in str(n)][::-1]

        # Double every other digit and add the others.  Modulo 10 on the result and it should equal 0
        return (sum(r[0::2]) + sum(sum(divmod(d*2,10)) for d in r[1::2])) % 10 == 0

    # Eliminate any non-numeric characters (such as hyphen and space) from the text
    if not text.isdigit():
        text = ''.join(i for i in text if i.isdigit())
    
    return luhn(text)

def _redact(text: str, replaceWith: str = '*') -> str:
    '''
    Redacts PAN to limit to just the first six and last four digits
    '''
    needsRejoined=False
    if not text.isdigit():
        # Identify the separators and their positions
        seps=dict()
        pos=0
        for c in text:
            if not c.isdigit():
                seps.update({pos: c})
                needsRejoined=True
            pos+=1

        # Rewrite the string to only include digits and then process 
        text=''.join([c for c in text if c.isdigit()])
    
    lastFourPos=len(text)-4
    firstSix=text[:6]
    middle=replaceWith*(lastFourPos-6)
    lastFour=text[lastFourPos:]
    result=firstSix+middle+lastFour
    
    if needsRejoined:
        for pos in seps.keys():
            result=result[:pos] + seps[pos] + result[pos:]

    return result
    

if __name__ == '__main__':
    testPANs=[
        '4012001037140001514E100010003220121800000011150',
        'but not this4893 0133 3538 6137or this',
        '3782-822463-10005',
        '371449635398431\nbutnot this',
        '3787 344936 71000',
        '345606077182423',
    ]

    results=[]

    for test in testPANs:
        match=findMatch(test)
        if match:
            results+=[match]

    print(results)


import multiprocessing as mp

import yaml

from piidigger import console
from piidigger import queuefuncs
from piidigger.globalvars import SENTINEL
from piidigger.logmanager import LogManager

def processResult(outFilename: str,
                  queue: mp.Queue,
                  stopEvent: mp.Event,
                  logManager: LogManager,):
    
    
    try:
        logger = logManager.getLogger('yaml_handler')
        logger.info('Starting YAML output processor (%s)', mp.current_process().pid)
        with open(outFilename, 'w', encoding="utf-8") as of:
            while True:
                if stopEvent.is_set():
                    break
                item = queuefuncs.getItem(queue)
                if item == SENTINEL:
                    break
                if item ==  None:
                    continue
                yaml.dump(item, of, indent=4)
    except KeyboardInterrupt:
        pass
    except PermissionError as e:
        console.error(str(e))
        stopEvent.set()
    finally:
        logger.info('Stopping %s (PID=%d)', mp.current_process().name, mp.current_process().pid)


'''Process DOCX files'''

import warnings
from collections.abc import Iterator

from docx2python import docx2python
from docx2python.iterators import iter_paragraphs

from piidigger.filehandlers._sharedfuncs import ContentHandler
from piidigger.globalvars import maxChunkSize
from piidigger.globalvars import defaultChunkCount
from piidigger.logmanager import LogManager

warnings.filterwarnings('ignore', category=UserWarning, module='docx2python')

# Each filehandler must have the following:
#   "handles" -     dictionary to identify lists of file extensions and mime types that the handler will manage.
#                   This will be read by globals upon initial load to build the full list of supported mime types and file extensions
#   "processFile" - Function that manages opening and reading of the file.  The main module will call this handler wtih the "processFile(filename)" function.
#                   processFile should provide the lines of text to each of the dataHandlers

handles={
    'ext': [
        '.docx',
    ],
    'mime': [
        'application/vnd.openxmlformats-officedocument.wordprocessingml.document',
        ],
}

def readFile(filename: str, 
             logManager: LogManager,
             maxChunkCount = defaultChunkCount) -> Iterator[str]:
    ''''
    Handle all file IO and text extraction operations for this file type.  Returns a list of results that have been validated by each datahandler.  
    "filename" is a string of the path and filename to process.  "handlers" is passed as a list of module objects that are called directly by processFile.
    '''

    logger = logManager.getLogger('docx_handler')
    
    try:
        # Read in all of the docx content and close the file
        docxContent=docx2python(filename)
        handler: ContentHandler = ContentHandler(maxContentSize = maxChunkSize * maxChunkCount)

        # This will iterate of the header, body and footer of the document, including all of the text and tables
        for line in iter_paragraphs(docxContent.document):
            handler.appendContent(line)
            if handler.contentBufferFull():
                yield handler.getContent()

        for comment in docxContent.comments:
            if not comment is None:
                handler.appendContent(comment[3])
                if handler.contentBufferFull():
                    yield handler.strip()

        # No size check -- we'll just append the properties to the end of the content and send it
        handler.appendContent(str(docxContent.core_properties))

        # Once we've processed the entire file, it's time to send that last bit of info that hasn't already been sent.
        logger.debug('%s: Read %d lines', filename, handler.totalBytes)

        # Return the last chunk of content    
        yield handler.finalizeContent()
        
    except FileNotFoundError:
        logger.error('%s: Previously discovered file no longer exists. File skipped', filename)
    except PermissionError as e:
        logger.error('%s: PermissionError.  File skipped.  Error message: %s', filename, str(e))
    except OSError as e:
        logger.error('%s: OSError.  File skipped.  Error message: %s', filename, str(e))
    except Exception as e:
        logger.error('%s: Unknown exception.  File skipped.  Error message: %s', filename, str(e))
    




import datetime
import multiprocessing as mp
import os
import pathlib
import platform
import string

import tomli

from piidigger import console
from piidigger import globalfuncs
from piidigger.getmime import testMagic
from piidigger.logmanager import LogManager

class File:
    def __init__(self, f: pathlib.Path, mimeType: str):
        self.filename=f.name
        self.path=f.parent
        self.ext=f.suffix
        self.mimeType=mimeType
        self.handler=globalfuncs.getFileHandlerName(self.ext, self.mimeType)
        self.times=(f.stat().st_atime, f.stat().st_mtime)
        self.size=f.stat().st_size
        
    def __lt__(self, other):
        return self.getFullPath() < other.getFullPath()

    def getFullPath(self):
        return os.path.join(self.path, self.filename)
    
    def getExtension(self):
        return self.ext
    
    def getOldAccessTime(self):
        return self.times[0]
    
    def getOldModTime(self):
        return self.times[1]
    
    def getFileHandlerName(self) -> str:
        return self.handler
    
    
    def getFileSize(self):
        return self.size
    
    def getTimeStamps(self) -> tuple:
        return self.times

class Config:
    def __init__(self, configFile: str, useDefault: bool=False,):
        
        if useDefault:
            console.normal('Using default configuration.')
            self.config=globalfuncs.getDefaultConfig()
        else:
            try:
                with open(configFile, 'rb') as file:
                    self.config = tomli.load(file)
            except FileNotFoundError:
                console.warn('Configuration file %s not found. Using default configuration.' % configFile)
                configFile = 'Internal Config'
                self.config=globalfuncs.getDefaultConfig()
            except tomli.TOMLDecodeError as e:
                console.error('Invalid configuration (%s)' % (configFile))
                console.error(str(e))
                exit(globalfuncs.errorCodes['invalidConfig'])
        
        self.config['rootPath']=str(pathlib.Path(os.getcwd()).absolute())
        self.config['maxProcs']=os.cpu_count()
        hostname=str(platform.node())
        self.config['hostname']=hostname
        timeStamp=datetime.datetime.now().strftime("%Y%m%d-%H%M%S")
        
        if _isAll(self.config['dataHandlers']):
            self.config['dataHandlers'] = globalfuncs.getSupportedDataHandlerNames()
        else:
            configHandlers=self.config['dataHandlers']
            notFound = [n for n in configHandlers if not n in globalfuncs.getSupportedDataHandlerNames()]
            if notFound:
                console.error("Unexpected data handler found in configuration file (%s)" % (configFile))
                console.error("The following data handlers will be ignored: " + str(notFound))
                self.config['dataHandlers'] = [n for n in configHandlers if n not in notFound]
        
        # Build the file names for each of the supported result file types
        outpath=self.config['results']['path']
        if not outpath.endswith('/'):
            outpath+='/'
        
        # We don't need this item now and it'll mess up the output file function if we leave it in.
        del self.config['results']['path']
        fileTypes=[('json', '.json'), 
                   ('text', '.txt')]

        for resultType, ext in fileTypes:
            filename=f"{outpath}{hostname}-{timeStamp}{ext}"
            # As set in the TOML file, the different results types are booleans.  We'll replace them with their pathname equiv if they were "enabled" to begin with.
            if self.config['results'][resultType]:
                self.config['results'][resultType]=filename
            else:
                del self.config['results'][type]

        # Fix any Windows paths replacing ALL with the actual drive letters and confirming that hard-coded starting directories exist
        if _isAll(self.config['includeFiles']['startDirs']['windows']):
            self.config['includeFiles']['startDirs']['windows'] = ['%s:\\' % d for d in string.ascii_uppercase if os.path.exists('%s:' % d)]        
        elif not all([os.path.isdir(d) for d in self.getStartDirs()]):
            console.error("Starting directory does not exist (%s). Check configuration file (%s)." % (self.getStartDirs(), configFile))
            exit(globalfuncs.errorCodes['invalidConfig'])


        # Replace "all" extensions with those currently supported by File Handlers
        if _isAll(self.config['includeFiles']['ext']):
            self.config['includeFiles']['ext'] = globalfuncs.getSupportedFileExts()
        else:
            # Before comparing to the expected list, fix the YAML-provided extensions if they don't start with a period
            configExts=[c if c.startswith('.') else '.' + c for c in self.config['includeFiles']['ext']]
            notFound = [n for n in configExts if not n in globalfuncs.getSupportedFileExts()]
            if notFound:
                console.error("Unexpected file extensions found in configuration file (%s)" % (configFile))
                console.error("The following file extensions will be ignored: " + str(notFound))
                self.config['includeFiles']['ext'] = [n for n in configExts if n not in notFound]
        
        # Replace "all" MIME types with those currently supported by File Handlers
        if testMagic():
            if _isAll(self.config['includeFiles']['mime']):
                self.config['includeFiles']['mime'] = globalfuncs.getSupportedFileMimes()
            else:
                configMimes=self.config['includeFiles']['mime']
                notFound = [n for n in configMimes if not n in globalfuncs.getSupportedFileMimes()]
                if notFound:
                    console.error("Unexpected MIME types found in configuration file (%s)" % (configFile))
                    console.error("The following MIME types will be ignored: " + str(notFound))
                    self.config['includeFiles']['mime'] = [n for n in configMimes if n not in notFound]
        else:
            self.config['includeFiles']['mime']=[]

        # Add Results and Log folders to the list of folders to exlude
        self.config['excludeDirs'][globalfuncs.getOSType()].append(str(pathlib.Path(self.getRootPath()) / outpath))
        self.config['excludeDirs'][globalfuncs.getOSType()].append(str(pathlib.Path(pathlib.Path(self.getRootPath()) / self.getLogFile()).parent))
 
    def getDataHandlers(self):
        return self.config['dataHandlers']
    
    def getEnabledOutputTypes(self):
        #return [key for key in self.config['results'].keys()]
        return self.config['results'].keys()
 
    def getExcludeDirs(self):
        return self.config['excludeDirs'][globalfuncs.getOSType()]
 
    def getFileExts(self):
        return self.config['includeFiles']['ext']
 
    def getConfig(self):
        return self.config
 
    def getLocalFilesOnly(self):
        return self.config['localFilesOnly']
    
    def getLogLevel(self):
        return self.config['logging']['logLevel']
 
    def getLogFile(self):
        return self.config['logging']['logFile']

    def getMaxFilesScanProcs(self):
        #return max(self.getMaxProcs() // 4, 1)
        return 1

    def getMaxProcs(self):
        return self.config['maxProcs']

    def getMimeTypes(self):
        return self.config['includeFiles']['mime']

    def getOutputFile(self, resultType: str = ""):
        if resultType:
            return self.config['results'][resultType]
        else:
            return self.config['results']

    def getRootPath(self):
        return self.config['rootPath']

    def getStartDirs(self):
        return self.config['includeFiles']['startDirs'][globalfuncs.getOSType()]
    
    def setMaxProcs(self, procs):
        self.config['maxProcs']=procs

class ProcessManager:
    def __init__(self, 
                 name: str,
                 logManager: LogManager,):
        
        self.processes: list = []
        self.name = name
        self.logger = logManager.getLogger(name=name,)
        
    def register(self, 
                 *,
                 target: callable,
                 name: str,
                 num_processes: int, 
                 args: tuple = None
                 ):
        
        p={
            'target': target,
            'start_order': len(self.processes) + 1,
            'shutdown_order': None,
            'name': name,
            'num_processes': num_processes,
            'processes': [],
            'started': False,
            'args': args,
            }
        
        self.processes.append(p)
        self.logger.info(f'{self.name}: Registered process {name} with {num_processes} processes.')

    def start(self):
        try:
            self.processes.sort(key=lambda x: x['start_order'])
            for i, process in enumerate(self.processes):
                if not process['started']:
                    process['shutdown_order'] = len(self.processes) - i
                    for j in range(process['num_processes']):
                        p = mp.Process(target=process['target'], 
                                    name=f'{process['name']}_{j}',
                                    args=process['args'],
                        )
                        process['processes'].append(p)
                        p.start()
                        self.logger.info(f'Started process {p.name} (PID={p.pid}).')
                    process['started'] = True
        except KeyboardInterrupt:
            self.terminate_all_processes()

    def wait_for_processes(self):
        try:
            self.processes.sort(key=lambda x: x['shutdown_order'],)
            for process in self.processes:
                for p in process['processes']:
                    self.logger.debug(f'Joining process {p.name} (PID={p.pid}).')
                    p.join()
        except KeyboardInterrupt:
            self.terminate_all_processes()

    def terminate_all_processes(self):
        self.logger.debug(f'{self.name}: Terminating all processes.')
        self.processes.sort(key=lambda x: x['shutdown_order'],)
        for process in self.processes:
            for p in process['processes']:
                self.logger.debug(f'Terminating process {p.name} (PID={p.pid}).')
                p.terminate()
                p.join()        

def _isAll(x) -> bool:
    '''
    Checks if the provided content is equal to the string 'all'.  If it's a list or a dictionary, it will compare only the first item.

    Returns None if it's one of those three types.
    '''

    s=''
    if isinstance(x, list):
        s = x[0]
    if isinstance(x, str):
        s= x
    if isinstance(x, dict):
        s = list(x.keys())[0]
    
    return s.lower() == 'all'


if __name__=="__main__":
    config=Config('piidigger.toml')

    print(config.getConfig())


import codecs
from collections.abc import Iterator

from piidigger.getencoding import getEncoding
from piidigger.filehandlers._sharedfuncs import ContentHandler
from piidigger.globalvars import maxChunkSize
from piidigger.globalvars import defaultChunkCount
from piidigger.logmanager import LogManager

# Each filehandler must have the following:
#   "handles" -     dictionary to identify lists of file extensions and mime types that the handler will manage.
#                   This will be read by globals upon initial load to build the full list of supported mime types and file extensions
#   "processFile" - Function that manages opening and reading of the file.  The main module will call this handler wtih the "processFile(filename)" function.
#                   processFile should provide the lines of text to each of the dataHandlers

handles={
    'ext': [
        '.aplt', '.applescript', '.armx', '.asp', '.asax', '.asmx', '.aspx',
        '.bat',
        '.c', '.cc', '.cfm', '.clj', '.cljs', '.clojure', '.cob', '.cpp', '.csh', '.csv',
        '.erl',
        '.h', '.hrl', '.htm', '.ht4', '.html', '.html5',
        '.go', '.gvy',
        '.j', '.json', '.js', '.jsp',
        '.log',
        '.perl', '.php', '.pl', '.ps1', '.py',
        '.rb',
        '.scpt', '.sdef', '.ser', '.sh',
        '.toml', '.txt',
        '.vb',
        '.xml',
        '.yaml',
        '.zsh',
    ],
    'mime': [
        'application/json',
        'application/toml',
        'application/xml',
        'text/html',
        'text/plain',
    ],
}


def readFile(filename: str, 
            logManager: LogManager,
            maxChunkCount: int = defaultChunkCount,
            ) -> Iterator[str]:
    ''''
    Handle all file IO and text extraction operations for this file type.  Returns a generator object tied to maxChunkSize (650) * maxChunkCount bytes of text.  
    "filename" is a string of the path and filename to process.
    '''

    logger = logManager.getLogger('plaintext_handler')
    
    enc = getEncoding(filename=filename, logManager=logManager)

    if enc == None:
        logger.info('%s: Unknown encoding type', filename)
        return ['']
    else:
        logger.debug('%s: Encoding %s', filename, enc)
    
    # After getting the encoding from chardet, replace any unexpected characters with a plain ASCII "?"
    # The risk is we could lose something important, but if that's the one piece of data anywhere on the file system that would have matched,
    # then it's a risk worth taking for a more stable discovery tool.  More likely is that we might miss ONE INSTANCE of data in a file system that has 
    # many more instances for discovery.

    # File IO is the bottle neck but we could also hit some really big files.
    # By returning (through yield) the content in chunks, we can strike a balance between memory consumption and file IO speed.

    # First we open the file, then we add each line so long as the resulting line length remains less than maxContentSize.  
    # For the last line, we add one word at a time until we reach the limit.  

    try:
        with codecs.open(filename, 'r', encoding=enc, errors='replace') as f:
            handler: ContentHandler = ContentHandler(maxContentSize = maxChunkSize * maxChunkCount)
            for line in f:
                handler.appendContent(line)
                if handler.contentBufferFull():
                    yield handler.getContent()

    # Once we've processed the entire file, it's time to send that last bit of info that hasn't already been sent.
        logger.debug('%s: Read %d lines', filename, handler.totalBytes)

        # Return the last chunk of content    
        yield handler.finalizeContent()

    except FileNotFoundError:
        logger.error('Previously discovered file no longer exists: %s. File skipped', f.absolute())
    except PermissionError as e:
        logger.error('PermissionError adding %s.  File skipped.  Error message: %s', f.absolute(), str(e))
    except OSError as e:
        logger.error('OSError adding %s.  File skipped.  Error message: %s', f.absolute(), str(e))
    except UnicodeDecodeError as e:
        logger.error('Unicode error processing file %s (enc=%s): %s', filename, enc, e)
    except LookupError as e:
        logger.error('Codec lookup error processing file %s (enc=%s)', filename, enc)
    except Exception as e:
        logger.error('Unknown exception on file %s.  File skipped.  Error message: %s', filename, str(e))


import multiprocessing as mp
import json

from piidigger import console
from piidigger import queuefuncs
from piidigger.globalvars import SENTINEL
from piidigger.logmanager import LogManager

def processResult(outFilename: str,
                  queue: mp.Queue,
                  stopEvent: mp.Event,
                  logManager: LogManager,
                 ):
    
    
    # For JSON output, we need to store all results in a list and write them once the queue is shutdown
    try:
        logger = logManager.getLogger('json_handler')
        logger.info('Starting JSON output processor (%s)', mp.current_process().pid)
        allResults=list()

        while True:
            if stopEvent.is_set():
                break
            item = queuefuncs.getItem(queue)
            if item == SENTINEL:
                break
            if item ==  None:
                continue
            allResults.append(item)
    except KeyboardInterrupt:
        pass
    finally:
        try:
            with open(outFilename, 'w', encoding='utf-8') as of:
                json.dump(allResults, of, indent=4)
            logger.info('Stopping %s (PID=%d)', mp.current_process().name, mp.current_process().pid)
        except PermissionError as e:
            console.error(str(e))
            stopEvent.set()


from multiprocessing import Queue
from queue import Empty
from time import sleep

TIMEOUT=.1

def clearQ(q: Queue):
    '''Clears a queue of all contents'''
    while True:
        try:
            _ = q.get(block=False)
        except Empty:
            break
        except KeyboardInterrupt:
            continue 
    
def getItem(q: Queue):
    '''Returns the next item in the queue.  Uses a timeout to prevent blocking indefinitely'''
    try:
        return q.get(timeout=TIMEOUT)
    except Empty:
        return None

def waitOnQ(q: Queue):
    while not q.empty():
        sleep(TIMEOUT)


from collections.abc import Iterator

import xlrd

from piidigger.filehandlers._sharedfuncs import ContentHandler
from piidigger.globalvars import excelBlankColLimit
from piidigger.globalvars import excelBlankRowLimit
from piidigger.globalvars import maxChunkSize
from piidigger.globalvars import defaultChunkCount
from piidigger.logmanager import LogManager

# Each filehandler must have the following:
#   "handles" -     dictionary to identify lists of file extensions and mime types that the handler will manage.
#                   This will be read by globals upon initial load to build the full list of supported mime types and file extensions
#   "processFile" - Function that manages opening and reading of the file.  The main module will call this handler wtih the "processFile(filename)" function.
#                   processFile should provide the lines of text to each of the dataHandlers

handles={
    'ext': [
        '.xls',
    ],
    'mime': [
        'application/vnd.ms-excel',
        'application/excel',
        ],
}

def readFile(filename: str, 
                logManager: LogManager,
                maxChunkCount = defaultChunkCount,
            ) -> Iterator[str]:
    ''''
    Handle all file IO and text extraction operations for this file type.  Returns a list of results that have been validated by each datahandler.  
    "filename" is a string of the path and filename to process.  "handlers" is passed as a list of module objects that are called directly by processFile.
    '''

    logger = logManager.getLogger('xls-handler')
    
    try:
        # Don't use "on_demand" in order to keep the code simpler.  All worksheets are loaded into RAM.

        book=xlrd.open_workbook(filename, on_demand=True, formatting_info=False,)
        logger.debug('%s: Read %d worksheets', filename, len(book.sheet_names()))
        for sheet in book.sheet_names():
            logger.debug('Processing worksheet: %s', str(sheet))
            activeSheet=book.sheet_by_name(sheet)
            handler: ContentHandler = ContentHandler(maxContentSize = maxChunkSize * maxChunkCount)
            blankRowCount=0
            rowCount=0
            totalRows=activeSheet.nrows
            totalCols=activeSheet.ncols
            # create a string with all of the content of this sheet
            # Iterate through each cell in each row.  If we reach a limit of blank cells, move to the next row
            # If we reach a limit of blank rows, then move to the next sheet.
            for row in range(totalRows):
                logger.debug('%s[Sheet %s]: Processing row [%d]', filename, sheet, rowCount)
                line: str = ''
                rowCount+=1
                rowHasData=False
                blankColCount=0
                for col in range(totalCols):
                    item=activeSheet.cell_value(row, col)
                    if item is None or item == '':
                        blankColCount+=1
                        if blankColCount > excelBlankColLimit:
                            break
                        continue
                    # xlrd converts all numbers to floats.  If the float is really an integer (ends in '.0'), convert it to a string without the decimal point
                    if type(item) == float and str(item)[-2:] == '.0':
                        item = str(item)[:-2]
                    line += str(item) + ' '
                    rowHasData=True
                handler.appendContent(line)
                if rowHasData:
                    blankRowCount=0
                else:
                    blankRowCount+=1
                    if blankRowCount > excelBlankRowLimit:
                        logger.debug('%s[Sheet %s]: Blank row count exceeded at row %d', filename, sheet, rowCount)
                        break
                if handler.contentBufferFull():
                    yield handler.getContent()
            book.unload_sheet(sheet)
            logger.debug('%s[Sheet %s]: Read content (%d bytes)', filename, sheet, handler.totalBytes)
            yield handler.finalizeContent()
            
        book.release_resources()
    except FileNotFoundError:
        logger.error('Previously discovered file no longer exists: %s. File skipped', filename)
    except PermissionError as e:
        logger.error('PermissionError adding %s.  File skipped.  Error message: %s', filename, str(e))
    except OSError as e:
        logger.error('OSError adding %s.  File skipped.  Error message: %s', filename, str(e))
    except xlrd.compdoc.CompDocError:
        logger.error('Excel document corrupted: %s', filename,)
    except xlrd.biffh.XLRDError:
        logger.error('Excel read error: %s', filename,)
    except Exception as e:
        logger.error('Unknown exception on file %s.  File skipped.  Error message: %s', filename, str(e))

    




errorCodes={
    "ok": 0,
    "userTerm": 1,
    "invalidConfig": 1,
    'unknownError': 1024,
}

defaultChunkCount=100_000
excelBlankRowLimit=250
excelBlankColLimit=500
maxChunkSize = 650
SENTINEL = '!!!STOPQUEUE!!!'


######################################################
############ Shared File Handler Functions ###########
######################################################

from collections import deque

class ContentHandler:
    def __init__(self, 
                 maxContentSize: int
                ):
        self.maxContentSize = maxContentSize
        self.contentBuffer = deque()
        self.bufferLength = 0
        self.totalBytes = 0
    
    def appendContent(self, line: str) -> None:
        '''Appends a line of text to the content buffer'''

        line = self.replaceChars(line)
        words = line.split()
        self.totalBytes += len(line)

        for word in words:
            self.contentBuffer.append(word)
            self.bufferLength += len(word) + 1
    
    def contentBufferFull(self) -> bool:
        '''Returns True if the content buffer is full, False otherwise'''

        return self.bufferLength >= self.maxContentSize

    def replaceChars(self, content: str) -> str:
        '''Replaces characters in a string to make the it simpler for data handler regexes.  
        Replaces newlines, carriage returns, and tabs with spaces.  Strips the string of leading and trailing whitespace.'''
        
        return content.replace('\n', ' ').replace('\r', ' ').replace('\t', ' ').strip()
    
    def getContent(self) -> str:
        '''Returns up to maxContentSize amount of data from the buffer'''

        content: list = []
        contentLength: int = 0

        while self.contentBuffer and contentLength < self.maxContentSize:
            word = self.contentBuffer.popleft()
            content.append(word)
            self.bufferLength -= len(word) + 1
            contentLength += len(word) + 1

        return ' '.join(content)
    
    def finalizeContent(self) -> str:
        '''Returns the remaining content in the buffer'''

        content = ' '.join(self.contentBuffer)
        self.bufferLength = 0
        self.contentBuffer = []
        
        return content



from queue import Queue

import pytest

from piidigger.filehandlers import docx
from piidigger.queuefuncs import clearQ
from piidigger.logmanager import LogManager

@pytest.mark.filehandlers
@pytest.mark.parametrize('filename, expected_result', [
                            ('testdata/docx/empty-file.docx', []),
                            ('testdata/docx/lorem-ipsum-1line-comments.docx', 
                                [
                                    ("Lorem ipsum dolor sit amet, consectetur adipiscing elit, sed do eiusmod tempor incididunt ut labore et dolore magna aliqua. COMMENT RESPONSE {'title': None, 'subject': None, 'creator': 'Randy Bartels', 'keywords': None, 'description': None, 'lastModifiedBy': 'Randy Bartels', 'revision': '4', 'created': '2024-03-28T21:21:00Z', 'modified': '2024-03-28T21:23:00Z'}", 363)
                                ]),
                            ('testdata/docx/lorem-ipsum-1line-header-footer.docx',
                                [
                                    ("Header Text Lorem ipsum dolor sit amet, consectetur adipiscing elit, sed do eiusmod tempor incididunt ut labore et dolore magna aliqua. Footer text {'title': None, 'subject': None, 'creator': 'Randy Bartels', 'keywords': None, 'description': None, 'lastModifiedBy': 'Randy Bartels', 'revision': '3', 'created': '2024-03-28T21:17:00Z', 'modified': '2024-04-03T19:21:00Z'}", 370)
                                ]),
                            ('testdata/docx/lorem-ipsum-1line-heading-toc.docx', 
                                [
                                    ("Contents Heading 1 1 Heading 1 Lorem ipsum dolor sit amet, consectetur adipiscing elit, sed do eiusmod tempor incididunt ut labore et dolore magna aliqua. {'title': None, 'subject': None, 'creator': 'Randy Bartels', 'keywords': None, 'description': None, 'lastModifiedBy': 'Randy Bartels', 'revision': '3', 'created': '2024-03-28T21:19:00Z', 'modified': '2024-03-28T21:21:00Z'}", 377),
                                ]),
                            ('testdata/docx/lorem-ipsum-1line-hyperlink.docx', 
                                [
                                    ("Lorem ipsum dolor sit amet, consectetur adipiscing elit, sed do eiusmod tempor incididunt ut labore et dolore magna aliqua. {'title': None, 'subject': None, 'creator': 'Randy Bartels', 'keywords': None, 'description': None, 'lastModifiedBy': 'Randy Bartels', 'revision': '2', 'created': '2024-03-28T21:25:00Z', 'modified': '2024-03-28T21:25:00Z'}", 346)
                                ]),
                            ('testdata/docx/lorem-ipsum-1line-with-footnote-endnote.docx', 
                                [
                                    ("Lorem ipsum dolor sit amet, consectetur adipiscing elit, sed do eiusmod tempor incididunt ut labore et dolore magna aliqua.----footnote1--------endnote1---- footnote1) FOOTNOTE endnote1) ENDNOTE {'title': None, 'subject': None, 'creator': 'Randy Bartels', 'keywords': None, 'description': None, 'lastModifiedBy': 'Randy Bartels', 'revision': '4', 'created': '2024-03-28T21:17:00Z', 'modified': '2024-03-28T21:18:00Z'}", 417)
                                ]),
                            ('testdata/docx/lorem-ipsum-1line-with-table.docx', 
                                [
                                    ("Lorem ipsum dolor sit amet, consectetur adipiscing elit, sed do eiusmod tempor incididunt ut labore et dolore magna aliqua. Cell R1C1 Cell R1C2 Cell R2C1 Cell R2C2 {'title': None, 'subject': None, 'creator': 'Randy Bartels', 'keywords': None, 'description': None, 'lastModifiedBy': 'Randy Bartels', 'revision': '2', 'created': '2024-03-28T21:16:00Z', 'modified': '2024-03-28T21:16:00Z'}", 386)
                                ]),
                            ('testdata/docx/lorem-ipsum-1line-wordart.docx', 
                                [
                                    ("Lorem ipsum dolor sit amet, consectetur adipiscing elit, sed do eiusmod tempor incididunt ut labore et dolore magna aliqua. WORDART WORDART {'title': None, 'subject': None, 'creator': 'Randy Bartels', 'keywords': None, 'description': None, 'lastModifiedBy': 'Randy Bartels', 'revision': '5', 'created': '2024-03-28T21:17:00Z', 'modified': '2024-03-29T14:31:00Z'}", 362)
                                ]),
                            ('testdata/docx/lorem-ipsum-1line.docx', 
                                [
                                    ("Lorem ipsum dolor sit amet, consectetur adipiscing elit, sed do eiusmod tempor incididunt ut labore et dolore magna aliqua. {'title': None, 'subject': None, 'creator': 'Randy Bartels', 'keywords': None, 'description': None, 'lastModifiedBy': 'Randy Bartels', 'revision': '2', 'created': '2024-03-28T21:17:00Z', 'modified': '2024-04-01T13:28:00Z'}", 346)
                                ]),
                            ('testdata/docx/lorem-ipsum-2paragraph.docx', 
                                [
                                    ('Lorem ipsum dolor sit amet, consectetur adipiscing elit, sed do eiusmod tempor incididunt ut labore et dolore magna aliqua. Fusce id velit ut tortor pretium viverra suspendisse potenti nullam. Eu sem integer vitae justo eget magna. Orci a scelerisque purus semper eget duis at tellus at. Tempor orci eu lobortis elementum nibh tellus molestie nunc. Aliquet lectus proin nibh nisl condimentum id. Eu tincidunt tortor aliquam nulla facilisi cras fermentum odio eu. Diam sollicitudin tempor id eu nisl nunc. Venenatis a condimentum vitae sapien pellentesque habitant morbi tristique senectus. Nunc faucibus a pellentesque sit amet porttitor eget. Dictum varius duis at consectetur lorem donec massa sapien faucibus. Amet venenatis urna cursus eget nunc scelerisque. Sit amet porttitor eget dolor morbi non arcu risus. Donec ultrices tincidunt arcu non sodales neque sodales. Tincidunt dui ut ornare lectus sit amet est placerat in. Cras pulvinar mattis nunc sed blandit libero volutpat. Sed cras ornare arcu dui vivamus arcu felis bibendum. Elementum facilisis leo vel fringilla est. Morbi enim nunc faucibus a pellentesque sit. Ipsum suspendisse ultrices gravida dictum. Urna nunc id cursus metus aliquam eleifend mi in. Amet consectetur adipiscing elit pellentesque. Dignissim cras tincidunt lobortis', 1299),
                                    ("feugiat vivamus at augue eget. Tristique et egestas quis ipsum suspendisse ultrices gravida dictum. Eu augue ut lectus arcu bibendum at varius vel. Eros donec ac odio tempor orci dapibus ultrices. Fermentum et sollicitudin ac orci phasellus. Magnis dis parturient montes nascetur ridiculus mus mauris. Integer quis auctor elit sed vulputate. Iaculis at erat pellentesque adipiscing. {'title': None, 'subject': None, 'creator': 'Randy Bartels', 'keywords': None, 'description': None, 'lastModifiedBy': 'Randy Bartels', 'revision': '2', 'created': '2024-03-28T21:14:00Z', 'modified': '2024-03-28T21:14:00Z'}", 605)
                                ]),
                            ('testdata/docx/does-not-exist.docx', []),
                          ]
                  )
def test_read_docx_file(filename, expected_result):
    # Cut back from the default of 100,000 to make testing a bit easier on super-large files.  This is based on the premise that if it can handle 2 chunks (1300 bytes), it can handle 100,000 (61MB).
    maxChunkCount=2
    logQ = Queue()
    logManager=LogManager(logFile='test.log', logLevel='INFO', logQueue=logQ)

    result: list = []

    for content in docx.readFile(filename, logManager, maxChunkCount):
        result.append((content, len(content)))

    print(f'Result: "{result}"')
    
    clearQ(logQ)

    assert result == expected_result


import argparse
import multiprocessing as mp
import sys
import textwrap
import traceback
from ctypes import c_int, c_uint64
from datetime import datetime
from os import makedirs, cpu_count
from pathlib import Path
from time import sleep


import json
from wakepy import keep


import piidigger.classes as classes
from piidigger import console
from piidigger import filescan
from piidigger import globalfuncs
from piidigger import queuefuncs
from piidigger import __version__
from piidigger.globalvars import errorCodes
from piidigger.globalvars import SENTINEL
from piidigger.logmanager import LogManager


def cleanup(queues: dict,):
    '''Cleans up all processes on ending the program'''

    # Clean up any queues that still have stuff in them.  This will ensure that the mp.Queue-based QueueHandler thread will also terminate
    print()
    for q in queues:
        if not queues[q].empty():
            queuefuncs.clearQ(queues[q])

def commandLineParser() -> argparse.ArgumentParser:
    '''Handles the command line options'''
    parser = argparse.ArgumentParser(
    prog=sys.argv[0].split('/')[-1],
    formatter_class=argparse.RawDescriptionHelpFormatter,
    description=textwrap.dedent('''
        Search the file system for Personally Identifiable Information

        NOTES:
            * All program configuration is kept in \'piidigger.toml\' -- a TOML-formatted configuration file
            * A default configuration will be used if the default 'piidigger.toml' file doesn't exist
        ''')
    )

    configControl = parser.add_argument_group(title='Configuration')
    configControl.add_argument(
        '-c', '--create-conf',
        dest='createConfigFile',
        default='',
        help='Create a default configuration file for editing/reuse.',
        )
    configControl.add_argument(
        '-d', '--default-conf',
        dest='defaultConfig',
        action='store_true',
        help='Use the default, internal config.'
        )
    configControl.add_argument(
        '-f', '--conf-file', 
        dest='configFile', 
        default='piidigger.toml', 
        help='path/to/configfile.toml configuration file (Default = "piidigger.toml").  If the file is not found, the default, internal configuration will be used.'
        )
    configControl.add_argument(
        '-p', '--max-process',
        dest='maxProc',
        default=0,
        type=int,
        help='Override the number processes to use for searching files.  Will use the lesser of CPU cores or this value.  On production servers, consider setting this to less than the number of physical CPUs.  See \'--cpu-count\' below.',
        )
    
    miscInfoControl = parser.add_argument_group(title='Misc. Info')
    miscInfoControl.add_argument(
        '--cpu-count',
        dest='cpuCount',
        action='store_true',
        help='Show the number of logical CPUs provided by the OS.  Use this to tune performance.  See \'--max-process\' above.'
    )
    miscInfoControl.add_argument(
        '--list-datahandlers',
        dest='listDH',
        action='store_true',
        help='Display the list of data handlers and exit'
    )
    miscInfoControl.add_argument(
        '--list-filetypes',
        dest='listFT',
        action='store_true',
        help='Display the list of file types and exit'
    )
    miscInfoControl.add_argument(
        '--version', '-v',
        dest='version',
        action='store_true',
        help='Display the version number and exit'
    )
    
    return parser.parse_args()


def fileHandlerDispatcher(config: classes.Config,
                          queues: dict,
                          totals: dict,
                          stopEvent: mp.Event,
                          activeFilesQProcesses: mp.Value,
                          logManager: LogManager,
                         ):

    try:
        with activeFilesQProcesses.get_lock():
            activeFilesQProcesses.value+=1
        dataHandlerModules=globalfuncs.getEnabledDataHandlerModules(config.getDataHandlers())

        logger = logManager.getLogger(name=mp.current_process().name)
        logger.debug('Process %s (%s) started (Active=%d)', mp.current_process().name, mp.current_process().pid, activeFilesQProcesses.value)
        resultsQs = [qName for qName in queues.keys() if qName.endswith('_resultsQ')]
        
        while True:
            if stopEvent.is_set():
                break
            item=queuefuncs.getItem(queues['filesQ'])
            if item == SENTINEL:
                break
            if item == None:
                continue

            # Set some variables for this item
            filename=item.getFullPath()
            fileHandlerModule=globalfuncs.getFileHandlerModule(item.getFileHandlerName())
            results={
                'filename': filename,
                'matches': {}
            }
            
            logger.info('[%s]Processing %s with %s', mp.current_process().name, filename, fileHandlerModule.__name__)
            
            for content in fileHandlerModule.readFile(filename, logManager):
                logger.debug('%s: Received %d bytes from file hander', filename, len(content))

                if content == '':
                    break
                
                # chunks=globalfuncs.makeChunks(content)
                # logger.debug('%s: Created %d chunks', filename, len(chunks))

                # We only need chunks from here out.  Conserve a little memory by deleting "content"
                #del content

                for handler in dataHandlerModules:
                    #logger.debug('%s: Processing %d chunks with %s', filename, len(chunks), handler.dhName)
                    # for chunk in chunks:
                    #     results=globalfuncs.processMatches(results, handler.findMatch(chunk), handler.dhName)
                    results=globalfuncs.processMatches(results, handler.findMatch(content), handler.dhName)
                    # logger.debug('%s: %d chunks processed with %s', filename, len(chunks), handler.dhName)

            # Update the status counters
            with totals['filesScanned'].get_lock():
                totals['filesScanned'].value+=1
            with totals['bytesScanned'].get_lock():
                totals['bytesScanned'].value+=item.getFileSize()
            
            # Submit the results to outputhandlers
            if len(results['matches']) > 0:
                logger.debug('%s: %s matches found', filename, str(results['matches'].keys()))
                
                # Since Python sets aren't serializable as a JSON object type, we'll convert our results to Lists now.
                logger.debug('%s: Rebuilding result sets into lists', filename)
                for handler in results['matches']:
                    for key in results['matches'][handler]:
                        l=list(results['matches'][handler][key])
                        results['matches'][handler][key]=l

                # Update the results totals
                with totals['totalResults'].get_lock():
                    totals['totalResults'].value += globalfuncs.countResults(results['matches'])
                for q in resultsQs:
                    queues[q].put(results)

            logger.debug('%s: Processing complete', filename)

    except KeyboardInterrupt:
        pass
    finally:
        logger.info('Stopping %s (PID=%d)', mp.current_process().name, mp.current_process().pid)
        with activeFilesQProcesses.get_lock():
            activeFilesQProcesses.value-=1
            logger.info('FileHandler processes remaining: %d', activeFilesQProcesses.value)
        if activeFilesQProcesses.value==0:
            # Put the sentinel on the results queues
            for q in [qName for qName in queues.keys() if qName.endswith('_resultsQ')]:
                queues[q].put(SENTINEL)
            
            # To allow the queue to shutdown properly, remove the last item from the filesQ if we're the last filesQ processor still running
            logger.info('[%s]Last FileHandler process terminated.  Clearing filesQ.', mp.current_process().name)
            queuefuncs.clearQ(queues['filesQ'])
        else:
            logger.info('[%s]FileHandler process terminated.  %d FileHandler processes remaining.', mp.current_process().name, activeFilesQProcesses.value)
            # Put another sentinel on the filesQ to signal the next fileHandler process to terminate
            queues['filesQ'].put(SENTINEL)
        del logger


def getOutputHandlers(config: classes.Config,
                      queues: dict,
                      stopEvent: mp.Event,
                      logManager: LogManager,):

    '''Setup the results queue handlers for the enabled output types.
    
    Returns a list of callables which will be added to the ProcessManager for execution.'''
    
    try:
        for resultsType in config.getEnabledOutputTypes():
            try:
                makedirs(str(Path(config.getOutputFile(resultsType)).absolute().parent), exist_ok=True)
            except Exception as e:
                console.error(str(e))
                stopEvent.set()
            yield {
                    'target': globalfuncs.getOutputHandlerModule(resultsType).processResult,
                    'name': resultsType+'_handler', 
                    'num_processes': 1,
                    'args': (config.getOutputFile(resultsType), 
                             queues[resultsType+'_resultsQ'], 
                             stopEvent,
                             logManager,),
                  }
    except KeyboardInterrupt:
        pass


def progressLineWorker(totals: dict, 
                       startTime: datetime,
                       stopEvent: mp.Event,
                       logManager: LogManager,
                      ):

    def _printProgressLine():
        INTERVAL = 1
        lastLineLen = 0

        while True:
            screenWidth=console.getTerminalSize()[0]
            line='{} | Folders scanned: {:,}/{:,} | Files scanned: {:,}/{:,} ({}/{}) | Results found: {}'.format(
                str(datetime.now() - startTime).split('.')[0], 
                totals['dirsScanned'].value, totals['dirsFound'].value, 
                totals['filesScanned'].value, totals['filesFound'].value, 
                globalfuncs.sizeof_fmt(totals['bytesScanned'].value), globalfuncs.sizeof_fmt(totals['bytesFound'].value),
                totals['totalResults'].value)
            if len(line) > screenWidth:
                line=line[:screenWidth-1]

            console.status(line+' '*(lastLineLen-len(line)))
            lastLineLen=len(line)

            #Placing the event handler at the end will ensure that the last update is processed before terminating the thread
            if stopEvent.is_set():
                break
            sleep(INTERVAL)

    try:
        logger = logManager.getLogger('progressLineWorker')
        logger.info('progressLineWorker started')

        console.normal('If needed, press CTRL-C to terminate scan')

        #_printProgressLine()

        # Removed temporarily as it was having trouble with PyInstaller  
        if globalfuncs.getOSType()=='linux':
            _printProgressLine()
        else:
            with keep.presenting() as k:
                if k.success:
                    console.normal('Sleep prevention enabled.')
                else:
                    console.warn('Sleep prevention was unsuccessful.  System may go to sleep during scan.')
                _printProgressLine()
    except KeyboardInterrupt:
        pass
    finally:
        logger.info('Stopping %s (PID=%d)', mp.current_process().name, mp.current_process().pid)

def main():
    start=datetime.now()
    args=commandLineParser()
    if args.cpuCount:
        print('CPU cores:', cpu_count())
        sys.exit(errorCodes['ok'])

    if args.listDH:
        print('Data handler modules: ', globalfuncs.getSupportedDataHandlerNames())
        sys.exit(errorCodes['ok'])

    if args.listFT:
        print('File extns: ', globalfuncs.getSupportedFileExts())
        print('MIME types: ', globalfuncs.getSupportedFileMimes())
        sys.exit(errorCodes['ok'])

    if args.version:
        print('PIIDigger version:', __version__)
        sys.exit(errorCodes['ok'])

    if len(args.createConfigFile) >0:
        tomlFile = str(args.createConfigFile) if str(args.createConfigFile).endswith('.toml') else str(args.createConfigFile)+'.toml'
        configFileWritten=globalfuncs.writeDefaultConfig(tomlFile)

        if configFileWritten=='Success':
            console.normal('Default configuration written to '+args.createConfigFile)
            sys.exit(errorCodes['ok'])
        else:
            console.error('Config file not written: %s' % (configFileWritten))
            sys.exit(errorCodes['unknown'])

    if args.defaultConfig:
        config=classes.Config(configFile='', useDefault=True)
    else:
        config=classes.Config(configFile=args.configFile)

    if args.maxProc>0:
        config.setMaxProcs(min(cpu_count(), args.maxProc))

    try:
        # Create queues and other structures needed for asynchronous implementation
        totals={k: mp.Value(c_uint64, 0) for k in [
            'dirsScanned', 
            'dirsFound', 
            'filesScanned', 
            'filesFound', 
            'bytesScanned', 
            'bytesFound',
            'totalResults']}
        queues={name: mp.Queue() for name in ['logQ', 'dirsQ', 'filesQ', 'totalsQ',]}
        activeFilesQProcesses=mp.Value(c_int, 0)
        for resultsType in config.getEnabledOutputTypes():
            name=resultsType+'_resultsQ'
            queues.update({name: mp.Queue()})
        stopEvent=mp.Event()
        stopEvent.clear()
        logManager=LogManager(
            logFile=config.getLogFile(), 
            logLevel=config.getLogLevel(), 
            logQueue=queues['logQ'],)
        
        # Configure logging / the logger process should be managed separately from all of the others.
        loggerPM=classes.ProcessManager(name='loggerPM', 
                                        logManager=logManager,)
        makedirs(str(Path(config.getLogFile()).absolute().parent), exist_ok=True)
        loggerPM.register(target=logManager.logProcessor, 
                    name='logProcessor',
                    num_processes=1, 
                    args=(stopEvent,))
        
        # We need to start the logProcessor process before we can start the other processes
        loggerPM.start()
        logger = logManager.getLogger('main')
            
        logger.info('Command line arguments: %s', sys.argv[1:])
        logger.info('Starting PIIDigger version %s', __version__)
        logger.info("Configuration: %s", json.dumps(config.getConfig()))

        if len(config.getMimeTypes()) == 0:        
            logger.info("MIME detection disabled.")

        if not globalfuncs.isAdmin():
            message='Not running as an administrator. File system access may be restricted.'
            console.warn(message)
            logger.info(message)

        if globalfuncs.getOSType() == 'linux':
            console.warn('Sleep prevention disabled on Linux. Consider using \'screen\' or \'tmux\' to ensure that PIIDigger survives an SSH disconnect.')
        
        console.normal('Scanning %s for files matching %s' % (config.getStartDirs(), config.getDataHandlers()))
        
        #####################################
        # Bring up the pipe line in reverse order -- starting with results and working back to scanning for directories
        #####################################
            
        # Setup each of the subprocesses that are needed
        mainPM=classes.ProcessManager(name='mainPM',
                                      logManager=logManager,)
        for outputHandler in getOutputHandlers(config, queues, stopEvent, logManager):
            mainPM.register(target=outputHandler['target'],
                        name=outputHandler['name'],
                        num_processes=outputHandler['num_processes'],
                        args=outputHandler['args'],)
        mainPM.register(target=filescan.findFilesWorker, 
                    name='findFilesWorker',
                    num_processes=config.getMaxFilesScanProcs(), 
                    args=(config, queues, totals, stopEvent, logManager,),)
        mainPM.register(target=fileHandlerDispatcher, 
                    name='fileHandler',
                    num_processes=config.getMaxProcs(),
                    args=(config, queues, totals, stopEvent, activeFilesQProcesses, logManager,),)
        mainPM.register(target=filescan.findDirsWorker, 
                    name='findDirsWorker',
                    num_processes=1,
                    args=(config, queues, totals, stopEvent, logManager,),)
        console.normal('Starting %d file scanner processes' % (config.getMaxFilesScanProcs()))
        console.normal('Starting %d file handler processes' % (config.getMaxProcs()))

        # Start the progress line worker in a separate process manager
        progressPM=classes.ProcessManager(name='progressPM',
                                          logManager=logManager,)
        progressPM.register(target=progressLineWorker, 
                    name='progressLineWorker',
                    num_processes=1, 
                    args=(totals, start, stopEvent, logManager,),)
        
        # Start the processes
        progressPM.start()
        mainPM.start()

        ##################################
        # Wait for the main program to finish
        ##################################

        mainPM.wait_for_processes()
    except KeyboardInterrupt:
        # If the keyboard interrupt was early enought, maybe not all of the process managers have been started
        # So wrap the cleanup in a try/except block to catch undeclared variables
        try:
            progressPM.terminate_all_processes()
            mainPM.terminate_all_processes()
            loggerPM.terminate_all_processes()
        except UnboundLocalError:
            pass
    except Exception:
        console.error('An unknown error was encountered.  Error message was captured in %s.' % config.getLogFile())
        logger.error(traceback.print_exc())
    else:
        queues['logQ'].put(SENTINEL)
        # If the logger hasn't already been shutdown by a KeyboardInterrupt or other event
        stopEvent.set()
        progressPM.wait_for_processes()
    finally:
        try:
            loggerPM.wait_for_processes()
            cleanup(queues)
        except UnboundLocalError:
            pass
        console.normal('Scan complete.')
            
if __name__ == '__main__':
    mp.freeze_support()
    main()


import sys
import traceback
from multiprocessing import Process, freeze_support, set_start_method

from piidigger.piidigger import main
from piidigger.globalvars import errorCodes

exitCode = errorCodes['ok']

if __name__=='__main__':
    freeze_support()
    try:
        m = Process(target=main)
        m.start()
        m.join()
        exitCode = m.exitcode
    except KeyboardInterrupt:
        pass
    except Exception:
        exitCode = errorCodes['unknownError']
        errorFile='piidigger.exc'
        print(f'An unknown error was encountered.  Detailed error information has been written to {errorFile}.')
        traceback.print_exception(file=errorFile)
    sys.exit(exitCode)


from logging import INFO

from chardet import UniversalDetector

from piidigger.logmanager import LogManager

def getEncoding(filename: str,
                logManager: LogManager,) -> str:
    '''
    Uses chardet to indenty the file encoding by reading MAXLINES of data.

    '''

    logger = logManager.getLogger('getEncoding')
    detector = UniversalDetector()
    detector.logger.level=INFO
        
    try:
        with open(filename, 'rb') as f:
            for line in f.readlines():
                detector.feed(line)
                if detector.done: 
                    break
        detector.close()
        guess=detector.result['encoding']
    except Exception as e:
        logger.info('%s: %s', filename, str(e))
        # Mimic the chardet.detector output to preserve code integrety for function consumers
        guess=None
    
    logger.debug('Filename %s chardet results: %s', filename, str(guess))

    return guess


import openpyxl
import warnings
from collections.abc import Iterator

from openpyxl.utils.exceptions import *
from zipfile import BadZipFile

from piidigger.filehandlers._sharedfuncs import ContentHandler
from piidigger.globalvars import excelBlankColLimit
from piidigger.globalvars import excelBlankRowLimit
from piidigger.globalvars import maxChunkSize
from piidigger.globalvars import defaultChunkCount
from piidigger.logmanager import LogManager

# Ignore the UserWarning message from OpenPyXL that seem to pop up here and there
warnings.filterwarnings('ignore', category=UserWarning, module='openpyxl')

# Each filehandler must have the following:
#   "handles" -     dictionary to identify lists of file extensions and mime types that the handler will manage.
#                   This will be read by globals upon initial load to build the full list of supported mime types and file extensions
#   "processFile" - Function that manages opening and reading of the file.  The main module will call this handler wtih the "processFile(filename)" function.
#                   processFile should provide the lines of text to each of the dataHandlers

handles={
    'ext': [
        '.xlsx',
        '.xlsm',
        '.xlst',
        '.xltm',
    ],
    'mime': [
        'application/vnd.openxmlformats-officedocument.spreadsheetml.sheet',
        'application/vnd.ms-excel.sheet.macroEnabled',
        'application/vnd.ms-excel.template',
        ],
}

def readFile(filename: str, 
             logManager: LogManager,
             maxChunkCount: int = defaultChunkCount,
            ) -> Iterator[str]:
    ''''
    Handle all file IO and text extraction operations for this file type.  Returns a list of results that have been validated by each datahandler.  
    "filename" is a string of the path and filename to process.  "handlers" is passed as a list of module objects that are called directly by processFile.
    '''

    logger = logManager.getLogger('xlsx_handler')

    content: str = ''
    totalBytes: int = 0
    maxContentSize = maxChunkSize * maxChunkCount

    
    try:
        # Don't use "on_demand" in order to keep the code simpler.  All worksheets are loaded into RAM.
        # Some spreadsheet dimensions can't be accurately determined -- e.g. if there's a lot of extraneous formatting to make it look "pretty"
        # We build a safety valve so that it stops after the first 5000 rows and 5000 columns.  If the "interesting data" is present outside of these limits...
        # well... it's probably not the ONLY instance of such data.

        book=openpyxl.load_workbook(filename=filename, read_only=True, data_only=True,)
        logger.debug('%s: Read %d worksheets', filename, len(book.sheetnames))
        for sheet in book.sheetnames:
            logger.debug('%s: Processing worksheet: %s', filename, str(sheet))
            activeSheet=book[sheet]
            handler: ContentHandler = ContentHandler(maxContentSize = maxChunkSize * maxChunkCount)
            blankRowCount=0
            rowCount=0
            # create a string with all of the content of this sheet
            # Iterate through each cell in each row.  If we reach a limit of blank cells, move to the next row
            # If we reach a limit of blank rows, then move to the next sheet.
            for row in activeSheet.iter_rows(values_only=True):
                rowCount+=1
                rowHasData=False
                line: str = ''
                blankColCount: int = 0
                for item in row:
                    if isinstance(item, openpyxl.cell.cell.MergedCell):
                        continue
                    if item is None or item == '':
                        blankColCount+=1
                        if blankColCount>excelBlankColLimit:
                            break
                        continue
                    line += str(item) + ' '
                    rowHasData=True
                handler.appendContent(line)

                if rowHasData:
                    blankRowCount=0
                else:
                    blankRowCount+=1
                    if blankRowCount>excelBlankRowLimit:
                        logger.debug('%s[Sheet %s]: Blank row count exceeded at row %d', filename, sheet, rowCount)
                        break
                if handler.contentBufferFull():
                    yield handler.getContent()
                    
            logger.debug('%s[Sheet %s]: Read content (%d bytes)', filename, sheet, len(content))
            yield handler.finalizeContent()
            
    except FileNotFoundError:
        logger.error('Previously discovered file no longer exists: %s. File skipped', filename)
    except PermissionError as e:
        logger.error('PermissionError adding %s.  File skipped.  Error message: %s', filename, str(e))
    except OSError as e:
        logger.error('OSError adding %s.  File skipped.  Error message: %s', filename, str(e))
    except CellCoordinatesException as e: 
        logger.error('%s: %s', filename, e)
    except IllegalCharacterError as e:
        logger.error('%s: %s', filename, e)
    except InvalidFileException as e:
        logger.error('%s: %s', filename, e)
    except SheetTitleException as e:
        logger.error('%s: %s', filename, e)
    except UserWarning as e:
        logger.error('%s: %s', filename, e)
    except BadZipFile as e:
        logger.error('%s: %s', filename, e)
    except Exception as e:
        logger.error('Unknown exception on file %s.  File skipped.  Error message: %s', filename, str(e))
    else:
        book.close()






from queue import Queue

import pytest

from piidigger.filehandlers import xls
from piidigger.queuefuncs import clearQ
from piidigger.logmanager import LogManager

@pytest.mark.filehandlers
@pytest.mark.parametrize('filename, expected_result', [
                            ('testdata/xls/empty-file.xls', [('', 0)]),
                            ('testdata/xls/random-data-table.xls', 
                                [
                                    ('First Name Last Name Gender Age Email Education Julia Montgomery Female 24 j.montgomery@randatmail.com Doctoral Amanda Bailey Female 30 a.bailey@randatmail.com Doctoral Lydia Brooks Female 24 l.brooks@randatmail.com Lower secondary Aldus Kelly Male 24 a.kelly@randatmail.com Bachelor John Thompson Male 30 j.thompson@randatmail.com Doctoral Ryan Douglas Male 24 r.douglas@randatmail.com Primary Savana Dixon Female 24 s.dixon@randatmail.com Lower secondary Ted Chapman Male 21 t.chapman@randatmail.com Primary Paul Andrews Male 26 p.andrews@randatmail.com Bachelor Albert Wright Male 30 a.wright@randatmail.com Master Martin Thomas Male 21 m.thomas@randatmail.com Upper secondary Tony Hunt Male 26 t.hunt@randatmail.com Lower secondary Adele Grant Female 30 a.grant@randatmail.com Upper secondary Mike Owens Male 27 m.owens@randatmail.com Primary Adison Higgins Male 29 a.higgins@randatmail.com Lower secondary James Cooper Male 18 j.cooper@randatmail.com Master Jessica Ross Female 24 j.ross@randatmail.com Upper secondary Abraham Andrews Male 23 a.andrews@randatmail.com Lower secondary Albert Campbell Male 20 a.campbell@randatmail.com Master Julian Alexander Male 20 j.alexander@randatmail.com Bachelor Aiden Smith Male 24 a.smith@randatmail.com Lower secondary Sam Alexander Male 21 s.alexander@randatmail.com', 1314),
                                    ('Lower secondary Chelsea Craig Female 20 c.craig@randatmail.com Upper secondary Sofia Cooper Female 26 s.cooper@randatmail.com Master Carl Nelson Male 19 c.nelson@randatmail.com Doctoral Annabella Cameron Female 30 a.cameron@randatmail.com Bachelor Sabrina Douglas Female 19 s.douglas@randatmail.com Doctoral Martin Cooper Male 28 m.cooper@randatmail.com Lower secondary Madaline Walker Female 22 m.walker@randatmail.com Doctoral Albert Stewart Male 28 a.stewart@randatmail.com Upper secondary Richard Ellis Male 21 r.ellis@randatmail.com Primary Albert Riley Male 18 a.riley@randatmail.com Lower secondary Isabella Moore Female 24 i.moore@randatmail.com Master Emily Sullivan Female 18 e.sullivan@randatmail.com Doctoral Adelaide Cameron Female 25 a.cameron@randatmail.com Upper secondary Sawyer Foster Male 29 s.foster@randatmail.com Master Blake Nelson Male 26 b.nelson@randatmail.com Doctoral Kristian Smith Male 20 k.smith@randatmail.com Master Sophia Ross Female 29 s.ross@randatmail.com Primary John Hawkins Male 24 j.hawkins@randatmail.com Master Justin Adams Male 29 j.adams@randatmail.com Lower secondary Victoria Richardson Female 18 v.richardson@randatmail.com Master George Gibson Male 27 g.gibson@randatmail.com Upper secondary Ned Perry Male 21 n.perry@randatmail.com Bachelor Amber Murray', 1303),
                                    ('Female 25 a.murray@randatmail.com Primary Edith Anderson Female 20 e.anderson@randatmail.com Primary Jordan Henderson Male 25 j.henderson@randatmail.com Upper secondary Lucas Harrison Male 26 l.harrison@randatmail.com Primary Sydney Riley Female 25 s.riley@randatmail.com Bachelor Dainton Barnes Male 30 d.barnes@randatmail.com Primary Tara Baker Female 26 t.baker@randatmail.com Doctoral Oscar Sullivan Male 25 o.sullivan@randatmail.com Primary Freddie Baker Male 20 f.baker@randatmail.com Master Alan Morris Male 28 a.morris@randatmail.com Upper secondary Vincent Barnes Male 21 v.barnes@randatmail.com Upper secondary Alisa Craig Female 23 a.craig@randatmail.com Upper secondary Adrianna Anderson Female 24 a.anderson@randatmail.com Doctoral Jordan Martin Male 23 j.martin@randatmail.com Lower secondary Garry Hill Male 19 g.hill@randatmail.com Primary Alisa Morris Female 21 a.morris@randatmail.com Bachelor Carl Barnes Male 24 c.barnes@randatmail.com Lower secondary Daryl Allen Male 21 d.allen@randatmail.com Master Aston Morris Male 21 a.morris@randatmail.com Doctoral Melanie Andrews Female 25 m.andrews@randatmail.com Upper secondary Elian Evans Male 28 e.evans@randatmail.com Primary Caroline Kelley Female 23 c.kelley@randatmail.com Doctoral Ryan Turner Male 25 r.turner@randatmail.com Doctoral', 1305),
                                    ('Andrew Ellis Male 29 a.ellis@randatmail.com Lower secondary Evelyn Kelly Female 24 e.kelly@randatmail.com Lower secondary Lyndon Andrews Male 25 l.andrews@randatmail.com Lower secondary Nicole Warren Female 20 n.warren@randatmail.com Lower secondary Roman Smith Male 19 r.smith@randatmail.com Lower secondary Savana Moore Female 28 s.moore@randatmail.com Upper secondary Sienna Wright Female 21 s.wright@randatmail.com Master Elise Murray Female 25 e.murray@randatmail.com Master Paige Russell Female 27 p.russell@randatmail.com Bachelor Adelaide Ferguson Female 25 a.ferguson@randatmail.com Upper secondary Myra Sullivan Female 25 m.sullivan@randatmail.com Lower secondary Dale Cunningham Male 19 d.cunningham@randatmail.com Doctoral Adison Armstrong Male 19 a.armstrong@randatmail.com Lower secondary Andrew Andrews Male 26 a.andrews@randatmail.com Lower secondary Tiana Riley Female 26 t.riley@randatmail.com Doctoral Amanda Taylor Female 19 a.taylor@randatmail.com Master Mike Campbell Male 23 m.campbell@randatmail.com Primary Elian Hall Male 27 e.hall@randatmail.com Primary Jordan Anderson Male 24 j.anderson@randatmail.com Bachelor Ned Holmes Male 20 n.holmes@randatmail.com Bachelor Sophia Hamilton Female 27 s.hamilton@randatmail.com Doctoral George Clark Male 23 g.clark@randatmail.com Bachelor', 1305),
                                    ('Darcy Evans Female 22 d.evans@randatmail.com Primary Thomas Nelson Male 30 t.nelson@randatmail.com Lower secondary Alissa Nelson Female 29 a.nelson@randatmail.com Lower secondary Maddie Barrett Female 20 m.barrett@randatmail.com Doctoral Ada Scott Female 30 a.scott@randatmail.com Primary Michael Baker Male 25 m.baker@randatmail.com Lower secondary Daisy Lloyd Female 26 d.lloyd@randatmail.com Master Kellan Gray Male 23 k.gray@randatmail.com Upper secondary Dominik Evans Male 27 d.evans@randatmail.com Primary Alan Alexander Male 22 a.alexander@randatmail.com Primary Ryan Warren Male 20 r.warren@randatmail.com Master Rafael Riley Male 21 r.riley@randatmail.com Master Catherine Morris Female 28 c.morris@randatmail.com Doctoral Darcy Gray Female 18 d.gray@randatmail.com Master Alina Morrison Female 23 a.morrison@randatmail.com Primary Walter Lloyd Male 24 w.lloyd@randatmail.com Lower secondary Alexia Johnson Female 19 a.johnson@randatmail.com Lower secondary Edith Wilson Female 30 e.wilson@randatmail.com Master Mary Craig Female 25 m.craig@randatmail.com Lower secondary Harold Howard Male 20 h.howard@randatmail.com Lower secondary Dexter Walker Male 25 d.walker@randatmail.com Bachelor Adrianna Grant Female 30 a.grant@randatmail.com Master Vanessa Mason Female 25 v.mason@randatmail.com', 1300),
                                    ('Master Brooke Campbell Female 26 b.campbell@randatmail.com Bachelor Adam Williams Male 23 a.williams@randatmail.com Lower secondary Camila Morris Female 22 c.morris@randatmail.com Upper secondary Carlos Johnston Male 26 c.johnston@randatmail.com Master Oscar Hunt Male 23 o.hunt@randatmail.com Lower secondary Melanie Williams Female 19 m.williams@randatmail.com Upper secondary Clark West Male 27 c.west@randatmail.com Bachelor Richard Barnes Male 28 r.barnes@randatmail.com Upper secondary Naomi Ryan Female 18 n.ryan@randatmail.com Doctoral Clark Harrison Male 26 c.harrison@randatmail.com Master Nicole Spencer Female 27 n.spencer@randatmail.com Upper secondary Alissa Henderson Female 23 a.henderson@randatmail.com Upper secondary Nicholas Bailey Male 30 n.bailey@randatmail.com Lower secondary Henry Stewart Male 18 h.stewart@randatmail.com Lower secondary Emily Williams Female 24 e.williams@randatmail.com Bachelor Charlotte Riley Female 28 c.riley@randatmail.com Bachelor Kristian Hill Male 23 k.hill@randatmail.com Upper secondary Honey Perry Female 27 h.perry@randatmail.com Lower secondary Frederick Hunt Male 20 f.hunt@randatmail.com Master George Harris Male 26 g.harris@randatmail.com Bachelor Rebecca Foster Female 23 r.foster@randatmail.com Doctoral Samantha Brooks Female 29 s.brooks@randatmail.com', 1316),
                                    ('Lower secondary Miranda Brown Female 30 m.brown@randatmail.com Upper secondary Cadie Clark Female 25 c.clark@randatmail.com Master Paige Wright Female 20 p.wright@randatmail.com Primary Reid Ferguson Male 20 r.ferguson@randatmail.com Lower secondary Eleanor Lloyd Female 23 e.lloyd@randatmail.com Master Adison Rogers Male 26 a.rogers@randatmail.com Doctoral Marcus Stevens Male 27 m.stevens@randatmail.com Bachelor Nicole Johnson Female 30 n.johnson@randatmail.com Doctoral Arnold Cameron Male 29 a.cameron@randatmail.com Master Catherine Baker Female 22 c.baker@randatmail.com Primary Alexander Richards Male 29 a.richards@randatmail.com Bachelor Daryl Miller Male 18 d.miller@randatmail.com Doctoral Victor Robinson Male 24 v.robinson@randatmail.com Doctoral Alfred Ferguson Male 24 a.ferguson@randatmail.com Primary Bruce Morgan Male 29 b.morgan@randatmail.com Bachelor Marcus West Male 18 m.west@randatmail.com Lower secondary Martin Foster Male 29 m.foster@randatmail.com Master Emma Lloyd Female 28 e.lloyd@randatmail.com Lower secondary Ashton Russell Male 21 a.russell@randatmail.com Master Brad Howard Male 23 b.howard@randatmail.com Upper secondary Blake Owens Male 24 b.owens@randatmail.com Doctoral Rubie Richardson Female 30 r.richardson@randatmail.com Lower secondary Ted Taylor Male 29', 1301),
                                    ('t.taylor@randatmail.com Master Paige Grant Female 26 p.grant@randatmail.com Lower secondary Sofia Stewart Female 24 s.stewart@randatmail.com Primary Abigail Henderson Female 29 a.henderson@randatmail.com Bachelor Brooke Reed Female 25 b.reed@randatmail.com Primary Brianna Ellis Female 29 b.ellis@randatmail.com Primary Chester Reed Male 25 c.reed@randatmail.com Lower secondary Emily Baker Female 25 e.baker@randatmail.com Primary Alisa Lloyd Female 28 a.lloyd@randatmail.com Bachelor Florrie Payne Female 21 f.payne@randatmail.com Master Sam Armstrong Male 25 s.armstrong@randatmail.com Upper secondary Edwin Hill Male 20 e.hill@randatmail.com Lower secondary Arianna Morris Female 26 a.morris@randatmail.com Bachelor Heather Nelson Female 29 h.nelson@randatmail.com Master Kelsey Turner Female 24 k.turner@randatmail.com Master Tyler Thomas Male 24 t.thomas@randatmail.com Upper secondary Connie Perkins Female 30 c.perkins@randatmail.com Bachelor Lyndon Murray Male 29 l.murray@randatmail.com Upper secondary Kimberly Harris Female 24 k.harris@randatmail.com Bachelor Abraham Kelley Male 22 a.kelley@randatmail.com Upper secondary Emma Warren Female 30 e.warren@randatmail.com Bachelor Lilianna Crawford Female 26 l.crawford@randatmail.com Bachelor Ted Watson Male 22 t.watson@randatmail.com Lower', 1301),
                                    ('secondary Fenton Foster Male 27 f.foster@randatmail.com Bachelor Cadie Cameron Female 21 c.cameron@randatmail.com Lower secondary Edgar Davis Male 20 e.davis@randatmail.com Lower secondary Mike Cooper Male 29 m.cooper@randatmail.com Bachelor Rebecca Andrews Female 29 r.andrews@randatmail.com Bachelor Amy Foster Female 19 a.foster@randatmail.com Lower secondary Annabella Martin Female 29 a.martin@randatmail.com Master Arianna Hamilton Female 30 a.hamilton@randatmail.com Lower secondary Miller Casey Male 22 m.casey@randatmail.com Master Gianna Hill Female 27 g.hill@randatmail.com Lower secondary Robert Cooper Male 20 r.cooper@randatmail.com Doctoral Penelope Morris Female 26 p.morris@randatmail.com Master Kristian Hawkins Male 21 k.hawkins@randatmail.com Upper secondary Spike Richardson Male 20 s.richardson@randatmail.com Master Adrianna Taylor Female 22 a.taylor@randatmail.com Bachelor John Stewart Male 30 j.stewart@randatmail.com Doctoral Rosie Warren Female 18 r.warren@randatmail.com Bachelor Alexander Jones Male 24 a.jones@randatmail.com Lower secondary Maria Higgins Female 29 m.higgins@randatmail.com Master Miller Smith Male 24 m.smith@randatmail.com Primary Garry Adams Male 29 g.adams@randatmail.com Upper secondary Martin Roberts Male 20 m.roberts@randatmail.com Primary Eleanor', 1302),
                                    ('Alexander Female 25 e.alexander@randatmail.com Master George Johnston Male 24 g.johnston@randatmail.com Upper secondary Sofia Nelson Female 21 s.nelson@randatmail.com Master Cherry Allen Female 18 c.allen@randatmail.com Master Jared Owens Male 24 j.owens@randatmail.com Doctoral Myra Douglas Female 26 m.douglas@randatmail.com Lower secondary Derek Stewart Male 30 d.stewart@randatmail.com Upper secondary Elise Cole Female 25 e.cole@randatmail.com Lower secondary Robert Ellis Male 19 r.ellis@randatmail.com Upper secondary Ned Johnson Male 28 n.johnson@randatmail.com Upper secondary Blake Elliott Male 22 b.elliott@randatmail.com Primary Samantha Stevens Female 22 s.stevens@randatmail.com Lower secondary Sienna Roberts Female 19 s.roberts@randatmail.com Doctoral Penelope Craig Female 19 p.craig@randatmail.com Upper secondary Tess Watson Female 19 t.watson@randatmail.com Bachelor Anna Hall Female 29 a.hall@randatmail.com Bachelor Daryl Richardson Male 18 d.richardson@randatmail.com Master Eddy Turner Male 20 e.turner@randatmail.com Doctoral Melissa Farrell Female 29 m.farrell@randatmail.com Doctoral Ada Craig Female 19 a.craig@randatmail.com Upper secondary Preston Thompson Male 28 p.thompson@randatmail.com Master Kirsten Barrett Female 21 k.barrett@randatmail.com Lower secondary Clark', 1300),
                                    ('Morrison Male 27 c.morrison@randatmail.com Bachelor Kristian Robinson Male 26 k.robinson@randatmail.com Lower secondary Mary Hill Female 25 m.hill@randatmail.com Upper secondary Roman Harper Male 19 r.harper@randatmail.com Primary James Morgan Male 27 j.morgan@randatmail.com Lower secondary Gianna Parker Female 18 g.parker@randatmail.com Master Frederick Morris Male 28 f.morris@randatmail.com Master Catherine Walker Female 20 c.walker@randatmail.com Doctoral Sawyer Walker Male 18 s.walker@randatmail.com Lower secondary Miley Cole Female 30 m.cole@randatmail.com Lower secondary John Fowler Male 25 j.fowler@randatmail.com Lower secondary Byron Martin Male 21 b.martin@randatmail.com Master Elise West Female 29 e.west@randatmail.com Bachelor Kristian Reed Male 23 k.reed@randatmail.com Bachelor Adelaide Hall Female 21 a.hall@randatmail.com Master Lilianna Howard Female 21 l.howard@randatmail.com Lower secondary Walter Hamilton Male 30 w.hamilton@randatmail.com Upper secondary Ellia Rogers Female 30 e.rogers@randatmail.com Bachelor Nicole Higgins Female 19 n.higgins@randatmail.com Doctoral Michelle Cameron Female 29 m.cameron@randatmail.com Lower secondary Joyce Morrison Female 29 j.morrison@randatmail.com Lower secondary Lilianna Thomas Female 18 l.thomas@randatmail.com Doctoral Marcus', 1301),
                                    ('Grant Male 21 m.grant@randatmail.com Master Dainton Allen Male 29 d.allen@randatmail.com Lower secondary Edwin Cunningham Male 30 e.cunningham@randatmail.com Primary Chloe Brooks Female 19 c.brooks@randatmail.com Primary Maddie Bailey Female 26 m.bailey@randatmail.com Upper secondary Sydney Hamilton Female 20 s.hamilton@randatmail.com Doctoral Maximilian Ryan Male 20 m.ryan@randatmail.com Lower secondary Nicole Adams Female 18 n.adams@randatmail.com Primary Luke Williams Male 30 l.williams@randatmail.com Upper secondary Amber Dixon Female 30 a.dixon@randatmail.com Upper secondary Mary Scott Female 30 m.scott@randatmail.com Bachelor Garry Murphy Male 29 g.murphy@randatmail.com Lower secondary Alfred Lloyd Male 29 a.lloyd@randatmail.com Master Reid Kelly Male 20 r.kelly@randatmail.com Doctoral Frederick Douglas Male 24 f.douglas@randatmail.com Lower secondary Stella Stevens Female 30 s.stevens@randatmail.com Lower secondary Sawyer Cooper Male 20 s.cooper@randatmail.com Master Briony Taylor Female 25 b.taylor@randatmail.com Bachelor Ryan Holmes Male 26 r.holmes@randatmail.com Primary Camila Perry Female 18 c.perry@randatmail.com Lower secondary Chelsea Brown Female 25 c.brown@randatmail.com Doctoral Paul Hunt Male 20 p.hunt@randatmail.com Primary Antony Morgan Male 28 a.morgan@randatmail.com', 1309),
                                    ('Primary Adam Murphy Male 29 a.murphy@randatmail.com Bachelor Roman Cooper Male 29 r.cooper@randatmail.com Primary Gianna Warren Female 19 g.warren@randatmail.com Master Brad Barnes Male 20 b.barnes@randatmail.com Primary Dexter Carter Male 18 d.carter@randatmail.com Doctoral Patrick Tucker Male 29 p.tucker@randatmail.com Master Sawyer Hawkins Male 24 s.hawkins@randatmail.com Master Alfred Payne Male 19 a.payne@randatmail.com Bachelor Aldus Wells Male 19 a.wells@randatmail.com Lower secondary Byron Wright Male 24 b.wright@randatmail.com Lower secondary Brad Wells Male 29 b.wells@randatmail.com Primary Fiona Elliott Female 25 f.elliott@randatmail.com Lower secondary Lilianna Morris Female 22 l.morris@randatmail.com Master Robert Scott Male 22 r.scott@randatmail.com Bachelor Antony Hill Male 30 a.hill@randatmail.com Bachelor Elise Davis Female 28 e.davis@randatmail.com Bachelor Maria Campbell Female 23 m.campbell@randatmail.com Primary Agata Taylor Female 21 a.taylor@randatmail.com Upper secondary Hailey Alexander Female 30 h.alexander@randatmail.com Master Honey Richardson Female 23 h.richardson@randatmail.com Lower secondary Abraham West Male 23 a.west@randatmail.com Upper secondary Chelsea Allen Female 30 c.allen@randatmail.com Upper secondary Stuart Cameron Male 25 s.cameron@randatmail.com', 1311),
                                    ('Master Lucas Clark Male 20 l.clark@randatmail.com Lower secondary Roman Harrison Male 19 r.harrison@randatmail.com Lower secondary Lucy Hall Female 18 l.hall@randatmail.com Bachelor Camila Morgan Female 23 c.morgan@randatmail.com Lower secondary Vincent Miller Male 25 v.miller@randatmail.com Doctoral Alexander Watson Male 24 a.watson@randatmail.com Bachelor Henry Sullivan Male 27 h.sullivan@randatmail.com Lower secondary Lucas Harper Male 23 l.harper@randatmail.com Upper secondary Lucy Lloyd Female 24 l.lloyd@randatmail.com Primary Amanda Baker Female 18 a.baker@randatmail.com Upper secondary Brad Richardson Male 25 b.richardson@randatmail.com Lower secondary Marcus Allen Male 30 m.allen@randatmail.com Bachelor Ada Wells Female 28 a.wells@randatmail.com Doctoral Aida Riley Female 28 a.riley@randatmail.com Upper secondary Elian Carter Male 30 e.carter@randatmail.com Master Daisy Hawkins Female 21 d.hawkins@randatmail.com Upper secondary Madaline Dixon Female 28 m.dixon@randatmail.com Doctoral Henry Davis Male 24 h.davis@randatmail.com Master Frederick Cameron Male 29 f.cameron@randatmail.com Bachelor Annabella Robinson Female 30 a.robinson@randatmail.com Upper secondary Kellan Harris Male 23 k.harris@randatmail.com Master Kevin Andrews Male 27 k.andrews@randatmail.com Doctoral Haris', 1302),
                                    ('Morgan Male 24 h.morgan@randatmail.com Master Daryl Spencer Male 29 d.spencer@randatmail.com Upper secondary Elian Rogers Male 18 e.rogers@randatmail.com Doctoral Miley Gray Female 23 m.gray@randatmail.com Lower secondary Maya Williams Female 26 m.williams@randatmail.com Master Alen Bailey Male 26 a.bailey@randatmail.com Primary Luke Davis Male 29 l.davis@randatmail.com Primary Edgar Cameron Male 24 e.cameron@randatmail.com Upper secondary Darcy Wilson Female 22 d.wilson@randatmail.com Upper secondary Reid Hall Male 23 r.hall@randatmail.com Primary Violet Alexander Female 27 v.alexander@randatmail.com Master Jasmine Henderson Female 18 j.henderson@randatmail.com Master Thomas Adams Male 23 t.adams@randatmail.com Master Paul Turner Male 22 p.turner@randatmail.com Master Cherry Richardson Female 20 c.richardson@randatmail.com Doctoral Rebecca Wilson Female 20 r.wilson@randatmail.com Doctoral Byron Stevens Male 18 b.stevens@randatmail.com Doctoral Jordan Kelly Male 27 j.kelly@randatmail.com Bachelor Clark Elliott Male 21 c.elliott@randatmail.com Upper secondary Bruce Turner Male 27 b.turner@randatmail.com Primary Vivian Morgan Female 18 v.morgan@randatmail.com Primary Lucia Richardson Female 19 l.richardson@randatmail.com Doctoral Violet Henderson Female 24 v.henderson@randatmail.com', 1301),
                                    ('Primary Violet Bailey Female 30 v.bailey@randatmail.com Doctoral Edgar Cooper Male 23 e.cooper@randatmail.com Doctoral Lily Hill Female 27 l.hill@randatmail.com Primary Valeria Johnson Female 21 v.johnson@randatmail.com Bachelor Emma Robinson Female 29 e.robinson@randatmail.com Primary Alford Spencer Male 22 a.spencer@randatmail.com Primary Blake Robinson Male 19 b.robinson@randatmail.com Lower secondary Emily Phillips Female 30 e.phillips@randatmail.com Primary Eddy Roberts Male 24 e.roberts@randatmail.com Doctoral Alina Allen Female 23 a.allen@randatmail.com Upper secondary Thomas Owens Male 28 t.owens@randatmail.com Bachelor Joyce Alexander Female 21 j.alexander@randatmail.com Bachelor Kristian Parker Male 24 k.parker@randatmail.com Primary Mary Perry Female 18 m.perry@randatmail.com Primary James Richards Male 24 j.richards@randatmail.com Bachelor Nicholas Roberts Male 20 n.roberts@randatmail.com Upper secondary Sofia Scott Female 19 s.scott@randatmail.com Upper secondary Alford Craig Male 29 a.craig@randatmail.com Bachelor Bruce Hawkins Male 25 b.hawkins@randatmail.com Lower secondary Daisy Myers Female 21 d.myers@randatmail.com Lower secondary Henry Cameron Male 21 h.cameron@randatmail.com Primary Martin Foster Male 20 m.foster@randatmail.com Lower secondary Freddie Dixon Male', 1303),
                                    ('23 f.dixon@randatmail.com Master Vanessa Grant Female 30 v.grant@randatmail.com Primary Alisa Lloyd Female 20 a.lloyd@randatmail.com Lower secondary Sawyer Kelly Male 24 s.kelly@randatmail.com Lower secondary Lily Phillips Female 29 l.phillips@randatmail.com Doctoral Tyler Lloyd Male 19 t.lloyd@randatmail.com Master Paul Spencer Male 27 p.spencer@randatmail.com Bachelor Savana Baker Female 22 s.baker@randatmail.com Upper secondary Tess Stevens Female 26 t.stevens@randatmail.com Lower secondary Samantha Owens Female 23 s.owens@randatmail.com Doctoral Byron Smith Male 19 b.smith@randatmail.com Doctoral Frederick Montgomery Male 25 f.montgomery@randatmail.com Bachelor Olivia Miller Female 27 o.miller@randatmail.com Bachelor Rubie Howard Female 29 r.howard@randatmail.com Lower secondary Adrianna Adams Female 20 a.adams@randatmail.com Lower secondary Oscar Andrews Male 20 o.andrews@randatmail.com Primary Adelaide Spencer Female 18 a.spencer@randatmail.com Master Connie Farrell Female 30 c.farrell@randatmail.com Upper secondary Emma Cole Female 26 e.cole@randatmail.com Bachelor Kirsten Riley Female 20 k.riley@randatmail.com Doctoral Maria Taylor Female 28 m.taylor@randatmail.com Master James Myers Male 29 j.myers@randatmail.com Bachelor Alfred Tucker Male 23 a.tucker@randatmail.com Upper', 1302),
                                    ('secondary Anna Bailey Female 28 a.bailey@randatmail.com Master Brooke Barrett Female 22 b.barrett@randatmail.com Doctoral Frederick Ryan Male 18 f.ryan@randatmail.com Bachelor Chelsea Ross Female 20 c.ross@randatmail.com Bachelor Bruce Scott Male 24 b.scott@randatmail.com Master Maximilian Clark Male 25 m.clark@randatmail.com Bachelor Ada Adams Female 24 a.adams@randatmail.com Doctoral Rosie Harper Female 19 r.harper@randatmail.com Primary Joyce Smith Female 25 j.smith@randatmail.com Upper secondary Wilson Hall Male 29 w.hall@randatmail.com Lower secondary Maya Baker Female 19 m.baker@randatmail.com Primary Honey Moore Female 21 h.moore@randatmail.com Upper secondary James Roberts Male 22 j.roberts@randatmail.com Primary Dale Barrett Male 24 d.barrett@randatmail.com Doctoral Sawyer Allen Male 25 s.allen@randatmail.com Bachelor Fenton Myers Male 28 f.myers@randatmail.com Bachelor Sarah Walker Female 27 s.walker@randatmail.com Doctoral Amanda Harper Female 19 a.harper@randatmail.com Master Eleanor Gray Female 19 e.gray@randatmail.com Doctoral Alina Ferguson Female 28 a.ferguson@randatmail.com Upper secondary Camila Davis Female 20 c.davis@randatmail.com Primary Anna Davis Female 27 a.davis@randatmail.com Bachelor Carlos Hawkins Male 26 c.hawkins@randatmail.com Bachelor Alfred Russell', 1302),
                                    ('Male 27 a.russell@randatmail.com Upper secondary Emily Watson Female 29 e.watson@randatmail.com Master Florrie Morrison Female 30 f.morrison@randatmail.com Lower secondary Rebecca Roberts Female 22 r.roberts@randatmail.com Doctoral Abraham Robinson Male 22 a.robinson@randatmail.com Doctoral Sawyer Miller Male 21 s.miller@randatmail.com Upper secondary Jessica Roberts Female 25 j.roberts@randatmail.com Lower secondary Edgar Kelley Male 18 e.kelley@randatmail.com Lower secondary Jordan Myers Male 29 j.myers@randatmail.com Primary Miranda Allen Female 27 m.allen@randatmail.com Bachelor Tyler Douglas Male 19 t.douglas@randatmail.com Primary Frederick Higgins Male 24 f.higgins@randatmail.com Lower secondary Deanna Perry Female 26 d.perry@randatmail.com Lower secondary Eric Russell Male 18 e.russell@randatmail.com Lower secondary Paige Stevens Female 24 p.stevens@randatmail.com Bachelor Preston Higgins Male 19 p.higgins@randatmail.com Lower secondary Brad Jones Male 26 b.jones@randatmail.com Upper secondary Violet Ellis Female 27 v.ellis@randatmail.com Bachelor Dexter Sullivan Male 29 d.sullivan@randatmail.com Doctoral Caroline Murray Female 28 c.murray@randatmail.com Upper secondary Rafael Russell Male 21 r.russell@randatmail.com Primary Anna Sullivan Female 21 a.sullivan@randatmail.com', 1302),
                                    ('Upper secondary Martin Douglas Male 29 m.douglas@randatmail.com Bachelor Chelsea Murray Female 23 c.murray@randatmail.com Primary Grace Robinson Female 30 g.robinson@randatmail.com Lower secondary Lydia Murphy Female 19 l.murphy@randatmail.com Upper secondary Fiona Reed Female 28 f.reed@randatmail.com Upper secondary Kirsten Spencer Female 21 k.spencer@randatmail.com Primary Roman Roberts Male 21 r.roberts@randatmail.com Doctoral April Andrews Female 25 a.andrews@randatmail.com Master Sofia Turner Female 28 s.turner@randatmail.com Lower secondary Sienna Ross Female 23 s.ross@randatmail.com Lower secondary Alissa Walker Female 30 a.walker@randatmail.com Lower secondary Honey Kelley Female 28 h.kelley@randatmail.com Doctoral Daisy Miller Female 29 d.miller@randatmail.com Bachelor Derek Phillips Male 26 d.phillips@randatmail.com Master Maximilian Myers Male 30 m.myers@randatmail.com Doctoral Stella Roberts Female 28 s.roberts@randatmail.com Lower secondary Arthur Owens Male 23 a.owens@randatmail.com Lower secondary Jacob Harris Male 30 j.harris@randatmail.com Upper secondary Reid Ryan Male 23 r.ryan@randatmail.com Master Valeria Cunningham Female 28 v.cunningham@randatmail.com Primary Stuart Reed Male 29 s.reed@randatmail.com Upper secondary Penelope Harris Female 27 p.harris@randatmail.com', 1308),
                                    ('Lower secondary Eddy Hall Male 23 e.hall@randatmail.com Bachelor Evelyn Hawkins Female 28 e.hawkins@randatmail.com Bachelor Miller Adams Male 25 m.adams@randatmail.com Primary Dale Rogers Male 26 d.rogers@randatmail.com Upper secondary Robert Elliott Male 30 r.elliott@randatmail.com Lower secondary Lilianna Mitchell Female 30 l.mitchell@randatmail.com Lower secondary Victoria Gibson Female 22 v.gibson@randatmail.com Bachelor Kristian Scott Male 25 k.scott@randatmail.com Lower secondary Amelia Anderson Female 27 a.anderson@randatmail.com Lower secondary Freddie Hill Male 29 f.hill@randatmail.com Upper secondary Chelsea Murphy Female 29 c.murphy@randatmail.com Primary Eddy Morrison Male 24 e.morrison@randatmail.com Doctoral Victoria Carroll Female 21 v.carroll@randatmail.com Bachelor Natalie Williams Female 29 n.williams@randatmail.com Primary Ned Elliott Male 30 n.elliott@randatmail.com Primary Reid Riley Male 20 r.riley@randatmail.com Lower secondary Adison Richardson Male 22 a.richardson@randatmail.com Lower secondary Sydney Evans Female 22 s.evans@randatmail.com Master Frederick Watson Male 21 f.watson@randatmail.com Primary David Harrison Male 28 d.harrison@randatmail.com Master Vanessa Cunningham Female 19 v.cunningham@randatmail.com Lower secondary Ellia Miller Female 20 e.miller@randatmail.com', 1320),
                                    ('Primary Lucas Allen Male 28 l.allen@randatmail.com Bachelor Maximilian Evans Male 24 m.evans@randatmail.com Doctoral Ned Harris Male 26 n.harris@randatmail.com Bachelor Carina Williams Female 23 c.williams@randatmail.com Doctoral Camila Gray Female 19 c.gray@randatmail.com Lower secondary Cadie Farrell Female 30 c.farrell@randatmail.com Lower secondary Dale Watson Male 23 d.watson@randatmail.com Bachelor Kevin Armstrong Male 22 k.armstrong@randatmail.com Doctoral Myra Ferguson Female 25 m.ferguson@randatmail.com Primary Kelsey Owens Female 28 k.owens@randatmail.com Lower secondary Annabella Jones Female 27 a.jones@randatmail.com Lower secondary Evelyn Richardson Female 26 e.richardson@randatmail.com Primary Andrew Hunt Male 27 a.hunt@randatmail.com Doctoral Brianna Allen Female 23 b.allen@randatmail.com Doctoral Nicole Henderson Female 28 n.henderson@randatmail.com Primary Fiona Crawford Female 22 f.crawford@randatmail.com Primary Jenna Scott Female 22 j.scott@randatmail.com Master Cherry Lloyd Female 27 c.lloyd@randatmail.com Primary Kelsey Moore Female 25 k.moore@randatmail.com Master Tara Ryan Female 23 t.ryan@randatmail.com Upper secondary Oscar Owens Male 28 o.owens@randatmail.com Primary Stuart Murphy Male 22 s.murphy@randatmail.com Lower secondary Tess Hawkins Female 24 t.hawkins@randatmail.com', 1322),
                                    ('Upper secondary Dainton Alexander Male 28 d.alexander@randatmail.com Master Kristian Barnes Male 23 k.barnes@randatmail.com Doctoral Amy Nelson Female 18 a.nelson@randatmail.com Lower secondary Lucas Barnes Male 25 l.barnes@randatmail.com Bachelor Agata Anderson Female 27 a.anderson@randatmail.com Primary Adam Allen Male 22 a.allen@randatmail.com Bachelor Naomi Bailey Female 23 n.bailey@randatmail.com Doctoral Charlotte Edwards Female 26 c.edwards@randatmail.com Lower secondary Eric Elliott Male 29 e.elliott@randatmail.com Lower secondary Sam Murphy Male 28 s.murphy@randatmail.com Lower secondary Samantha Murray Female 26 s.murray@randatmail.com Upper secondary Alan Fowler Male 29 a.fowler@randatmail.com Bachelor Maria Stevens Female 18 m.stevens@randatmail.com Master Edwin Anderson Male 25 e.anderson@randatmail.com Master Dale Parker Male 29 d.parker@randatmail.com Lower secondary Stuart Grant Male 28 s.grant@randatmail.com Master Heather Hunt Female 21 h.hunt@randatmail.com Upper secondary Preston Cooper Male 21 p.cooper@randatmail.com Master Belinda Brown Female 25 b.brown@randatmail.com Upper secondary Ryan Campbell Male 30 r.campbell@randatmail.com Lower secondary Lucia Carter Female 28 l.carter@randatmail.com Lower secondary Emma Mason Female 30 e.mason@randatmail.com Doctoral', 1303),
                                    ('Tyler Thompson Male 22 t.thompson@randatmail.com Doctoral Fiona Robinson Female 19 f.robinson@randatmail.com Upper secondary Catherine Douglas Female 19 c.douglas@randatmail.com Lower secondary Samantha Thomas Female 26 s.thomas@randatmail.com Doctoral Alberta Clark Female 19 a.clark@randatmail.com Lower secondary Kelvin Barrett Male 21 k.barrett@randatmail.com Bachelor Amanda Douglas Female 23 a.douglas@randatmail.com Doctoral Ryan Warren Male 27 r.warren@randatmail.com Primary Walter Higgins Male 21 w.higgins@randatmail.com Primary Tara Richards Female 30 t.richards@randatmail.com Doctoral Leonardo Hawkins Male 24 l.hawkins@randatmail.com Primary Aiden Johnston Male 29 a.johnston@randatmail.com Lower secondary Vivian Cooper Female 28 v.cooper@randatmail.com Primary Elise Montgomery Female 24 e.montgomery@randatmail.com Primary Ellia Taylor Female 23 e.taylor@randatmail.com Primary Clark Miller Male 30 c.miller@randatmail.com Bachelor Connie Hamilton Female 22 c.hamilton@randatmail.com Primary Ted Phillips Male 30 t.phillips@randatmail.com Doctoral Carlos Howard Male 21 c.howard@randatmail.com Bachelor Kirsten Thompson Female 27 k.thompson@randatmail.com Upper secondary Eric Andrews Male 29 e.andrews@randatmail.com Master Freddie Clark Male 29 f.clark@randatmail.com Doctoral Jasmine', 1304),
                                    ('Hawkins Female 18 j.hawkins@randatmail.com Lower secondary Fiona Stevens Female 22 f.stevens@randatmail.com Primary Paige Crawford Female 21 p.crawford@randatmail.com Doctoral Jacob Stevens Male 23 j.stevens@randatmail.com Bachelor Dominik Spencer Male 20 d.spencer@randatmail.com Doctoral Lily Thompson Female 22 l.thompson@randatmail.com Upper secondary Florrie Richardson Female 22 f.richardson@randatmail.com Doctoral Dainton Warren Male 26 d.warren@randatmail.com Lower secondary Jack Stevens Male 22 j.stevens@randatmail.com Master Lydia Stevens Female 25 l.stevens@randatmail.com Upper secondary Lana Cooper Female 27 l.cooper@randatmail.com Doctoral Heather Phillips Female 23 h.phillips@randatmail.com Master Brianna Carroll Female 27 b.carroll@randatmail.com Doctoral Rafael Hamilton Male 26 r.hamilton@randatmail.com Doctoral Kirsten Williams Female 30 k.williams@randatmail.com Bachelor Dale Smith Male 30 d.smith@randatmail.com Master Maria Clark Female 23 m.clark@randatmail.com Primary Savana Stewart Female 28 s.stewart@randatmail.com Master Leonardo Holmes Male 22 l.holmes@randatmail.com Master Leonardo Baker Male 23 l.baker@randatmail.com Doctoral Briony Hunt Female 28 b.hunt@randatmail.com Primary Anna Holmes Female 30 a.holmes@randatmail.com Primary Adam Ellis Male 18 a.ellis@randatmail.com', 1315),
                                    ('Lower secondary Adrian Harrison Male 21 a.harrison@randatmail.com Doctoral Walter Holmes Male 27 w.holmes@randatmail.com Primary Robert Wilson Male 25 r.wilson@randatmail.com Primary Alexia Russell Female 26 a.russell@randatmail.com Lower secondary Kirsten Nelson Female 24 k.nelson@randatmail.com Master Jacob Montgomery Male 22 j.montgomery@randatmail.com Bachelor Grace Nelson Female 30 g.nelson@randatmail.com Upper secondary Adrianna Evans Female 30 a.evans@randatmail.com Bachelor Alisa Clark Female 26 a.clark@randatmail.com Lower secondary Nicole Cooper Female 25 n.cooper@randatmail.com Primary Catherine Wells Female 30 c.wells@randatmail.com Master Paul Rogers Male 20 p.rogers@randatmail.com Lower secondary Haris Ryan Male 22 h.ryan@randatmail.com Bachelor Camila Harrison Female 22 c.harrison@randatmail.com Bachelor Belinda Richardson Female 20 b.richardson@randatmail.com Upper secondary Alexia Harper Female 28 a.harper@randatmail.com Upper secondary Vivian Owens Female 23 v.owens@randatmail.com Bachelor Roman Owens Male 23 r.owens@randatmail.com Master Alisa Richardson Female 18 a.richardson@randatmail.com Primary Valeria Carter Female 25 v.carter@randatmail.com Primary Tony Payne Male 29 t.payne@randatmail.com Doctoral Oliver Payne Male 19 o.payne@randatmail.com Lower secondary', 1303),
                                    ('Samantha Watson Female 23 s.watson@randatmail.com Upper secondary David Warren Male 21 d.warren@randatmail.com Bachelor Camila Fowler Female 25 c.fowler@randatmail.com Upper secondary Sawyer Richards Male 30 s.richards@randatmail.com Master Kristian Chapman Male 22 k.chapman@randatmail.com Primary Leonardo Payne Male 25 l.payne@randatmail.com Master Madaline Russell Female 19 m.russell@randatmail.com Primary Elian Cameron Male 21 e.cameron@randatmail.com Doctoral Dale Hunt Male 18 d.hunt@randatmail.com Primary Mike Evans Male 20 m.evans@randatmail.com Primary Eleanor Casey Female 26 e.casey@randatmail.com Doctoral John Thompson Male 22 j.thompson@randatmail.com Upper secondary Emily Taylor Female 23 e.taylor@randatmail.com Lower secondary Michael Davis Male 22 m.davis@randatmail.com Bachelor Lyndon Kelly Male 30 l.kelly@randatmail.com Bachelor Joyce Dixon Female 19 j.dixon@randatmail.com Bachelor Adison Myers Male 30 a.myers@randatmail.com Doctoral Brooke Harper Female 18 b.harper@randatmail.com Lower secondary', 1026),
                                ]),
                            ('testdata/xls/test-1sheet-1cell.xls', [('S1R1C1', 6)]),
                            ('testdata/xls/test-1sheet-1row.xls', [("S1R1C1 S1R1C2 S1R1C3 S1R1C4", 27)]),
                            ('testdata/xls/test-1sheet-10row-table.xls', 
                                [
                                    ('Sheet Row Column Text 1 1 3 S1R1C3 1 2 3 S1R2C3 1 3 3 S1R3C3 1 4 3 S1R4C3 1 5 3 S1R5C3 1 6 3 S1R6C3 1 7 3 S1R7C3 1 8 3 S1R8C3 1 9 3 S1R9C3 1 10 3 S1R10C3', 153),
                                ]),
                            ('testdata/xls/test-2sheet-10row-table.xls', 
                                [
                                    ('Sheet Row Column Text 1 1 3 S1R1C3 1 2 3 S1R2C3 1 3 3 S1R3C3 1 4 3 S1R4C3 1 5 3 S1R5C3 1 6 3 S1R6C3 1 7 3 S1R7C3 1 8 3 S1R8C3 1 9 3 S1R9C3 1 10 3 S1R10C3', 153),
                                    ('Sheet Row Column Text 1 1 3 S1R1C3 1 2 3 S1R2C3 1 3 3 S1R3C3 1 4 3 S1R4C3 1 5 3 S1R5C3 1 6 3 S1R6C3 1 7 3 S1R7C3 1 8 3 S1R8C3 1 9 3 S1R9C3 1 10 3 S1R10C3', 153),
                            ]),
                            ('testdata/xls/does-not-exist.xls', []),
                          ]
                  )
def test_read_xls_file(filename, expected_result):
    # Cut back from the default of 100,000 to make testing a bit easier on super-large files.  This is based on the premise that if it can handle 2 chunks (1300 bytes), it can handle 100,000 (61MB).
    maxChunkCount=2
    logQ = Queue()
    logManager=LogManager(logFile='test.log', logLevel='INFO', logQueue=Queue())

    result: list = []

    for content in xls.readFile(filename, logManager, maxChunkCount):
        result.append((content, len(content)))

    print(f'Result: "{result}"')
        

    clearQ(logQ)

    assert result == expected_result


import logging
import multiprocessing as mp
from logging.handlers import QueueHandler
from time import sleep

from piidigger import (
    console,
    queuefuncs,)
from piidigger.globalvars import SENTINEL

class LogManager:
    def __init__(self,
                 logFile: str,
                 logLevel: str,
                 logQueue: mp.Queue,) -> None:
        
        self.logFile=logFile
        self.logLevel=logLevel
        self.logQueue=logQueue

    def logProcessor(self, stopEvent: mp.Event):
        """
        Processes log events from a queue and writes them to a log file.
        """
        try:
            logger = logging.getLogger('logProcessor')

            logFileFormatter=logging.Formatter('%(asctime)s:[%(name)s]:%(levelname)s:%(message)s')
            logFileHandler=logging.FileHandler(filename=self.logFile,mode='w',encoding='utf-8')
            logFileHandler.setFormatter(logFileFormatter)

            logger.setLevel(self.logLevel)
            logger.addHandler(logFileHandler)

            logger.info('Starting logProcessor (%s)', mp.current_process().pid)
            stopCause = None

            while True:
                if stopEvent.is_set():
                    stopCause = 'stopEvent'
                    break

                message = queuefuncs.getItem(self.logQueue)

                if message == SENTINEL:
                    stopCause = 'endQueue'
                    break

                if message is None:
                    continue

                logger.handle(message)
        except KeyboardInterrupt:
            console.normal('\n')
            console.warn('User terminated scan.  Shutting down.')

            # Give other processes a chance to write their final messages to the queue
            sleep(2)
            queuefuncs.clearQ(self.logQueue)
            stopCause='ctrlc'
        finally:
            logger.info('Stopping logProcessor [%s] (PID=%d)', str(stopCause), mp.current_process().pid)

    def getLogger(self,
                  name: str = '',) -> logging.Logger:
        """
        Returns a logger instance for subprocesses to use, which puts log events onto a queue.
        """
        logger = logging.getLogger(name)
        logger.setLevel(self.logLevel)
        if not logger.handlers:
            logger.addHandler(QueueHandler(self.logQueue))
        logger.propagate = False

        return logger


'''This data handler is not very well developed and in the interest of getting PIIDigger released to find PAN, it was more or less abandoned.  

The whole thing can probably be replaced with something built around this: https://stackabuse.com/validate-email-addresses-in-python-with-email-validator/

I don't recommend using it for now.'''

import re

dhName='Email Address'

def findMatch(line: str) -> dict:
    '''
    Matches a line of text against email addresses.  Should receive the text from "filehandler" as raw text.

    Returns a dictionary of:
        'type': set(matches)
    '''

    # Regexes provided by:
    #   email: https://stackabuse.com/python-validate-email-address-with-regular-expressions-regex/

    # But this one might be better? https://stackoverflow.com/questions/201323/how-can-i-validate-an-email-address-using-a-regular-expression 

    # If you know of more, or can shed light on any corrections, please submit an issue at https://github.com/kirkpatrickprice/PIIDigger/issues or submit a PR on the repo

    regexes={
        'email': re.compile(r"(([-!#-'*+/-9=?A-Z^-~]+(\.[-!#-'*+/-9=?A-Z^-~]+)*|\"([]!#-[^-~ \t]|(\\[\t -~]))+\")@([-!#-'*+/-9=?A-Z^-~]+(\.[-!#-'*+/-9=?A-Z^-~]+)*|\[[\t -Z^-~]*]))"),
        }
    
    results=dict()
    for datatype in regexes.keys():
        matches=regexes[datatype].findall(line)
        if matches:
            for match in matches:
                matchedText=match[0]
                matchedText=matchedText.strip()
                results[datatype]=set()
                redacted=_redact(matchedText, datatype)
                try:
                    results[datatype].add(redacted)
                except KeyError:
                    results[datatype]=[]
                    results[datatype]+=[_redact(redacted, datatype)]
    return results

def _redact(text: str, datatype: str, replaceWith: str = '*') -> str:
    
    def email() -> str:
        '''
        Redacts the email address
        '''

        parts=text.split('@')
        pii=parts[0]
        domain=parts[1]
        keepLength=1
        parts=list()
        parts.append(pii[0:keepLength])
        parts.append(replaceWith*(len(pii)-(keepLength*2)))
        parts.append(pii[-keepLength:])
        parts.append('@')
        parts.append(domain)
        
        return ''.join(parts)
    
    def phone_us():
        '''
        Redacts US-based phone numbers
        '''

        lastFour=replaceWith*4
        return text[:-4]+lastFour
    
    
    return locals()[datatype]()
    

if __name__ == '__main__':
    testdata=[
        'someone@example.com',
        'another.person@nowhere.co.uk',
        '+1 512 555 1234',
        '(512) 555-1234 is also a phone number',
    
    ]

    results=[]

    for test in testdata:
        match=findMatch(test)
        if match:
            results+=[match]

    print(results)